from __future__ import annotations

import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, List, Union

from rich.console import Console, RenderableType
from rich.markup import escape
from rich.text import Text
from rich.theme import Theme
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from cappa.command import FinalCommand

__all__ = [
    "Displayable",
    "Exit",
    "HelpExit",
    "Output",
    "error_format",
    "error_format_without_short_help",
    "output_format",
    "theme",
]


Displayable: TypeAlias = RenderableType
Outputable: TypeAlias = Union[List[Displayable], Displayable, "Exit", str, Any, None]


class Exit(SystemExit):
    def __init__(
        self,
        message: list[Displayable] | Displayable | None = None,
        *,
        command: FinalCommand[Any] | None = None,
        code: str | int | None = 0,
        prog: str | None = None,
    ):
        self.message = message
        self.prog = prog
        self.command = command
        super().__init__(code)


class HelpExit(Exit):
    def __init__(
        self,
        message: list[Displayable] | Displayable,
        *,
        code: str | int | None = 0,
        prog: str | None = None,
    ):
        super().__init__(code=code, prog=prog)
        self.message = message


theme: Theme = Theme(
    {
        "cappa.prog": "grey50",
        "cappa.group": "dark_orange bold",
        "cappa.arg": "cyan",
        "cappa.arg.name": "dark_cyan",
        "cappa.subcommand": "dark_cyan",
        "cappa.help": "",
    }
)


output_format: str = "{message}"
error_format: str = "{short_help}\n\n[red]Error[/red]: {message}"
error_format_without_short_help: str = "[red]Error[/red]: {message}"


@dataclass
class Output:
    """Output sink for CLI std out and error streams.

    For simple customization (namely disabling color and overriding the theme),
    `invoke` and `parse` accept `color` and `theme` arguments, which internally
    configure the `output_console` and `error_console` fields.

    For more involved customization, an Output object can be supplied into either `invoke`
    or `parse` functions as well.

    Note, all input arguments to Output are optional.

    Arguments:
        output_console: Output sink, defaults to printing to stdout.
        error_console: Error sink, defaults to printing to stderr.
        output_format: Format string through which output_console output will be
            formatted. The following format string named format arguments can be used
            in `output_format`: prog, code, message.
        error_format: Format string through which output_console error will be
            formatted. The following format string named format arguments can be used
            in `error_format`: prog, code, message, help, short_help.

    Examples:
        >>> output = Output()
        >>> output = Output(error_format="{prog}: error: {message}")
    """

    output_console: Console = field(
        default_factory=lambda: Console(file=sys.stdout, theme=theme)
    )
    error_console: Console = field(
        default_factory=lambda: Console(file=sys.stderr, theme=theme)
    )

    output_format: str = output_format
    error_format: str = error_format

    def color(self, value: bool = True):
        """Override the default `color` setting (None), to an explicit True/False value."""
        self.output_console.no_color = not value
        self.error_console.no_color = not value
        return self

    def theme(self, t: Theme | None):
        """Override the default Theme, or reset the theme back to default (with `None`)."""
        self.output_console.push_theme(t or theme)
        self.error_console.push_theme(t or theme)

    def exit(
        self,
        e: Exit,
        help: list[Displayable] | None = None,
        short_help: Displayable | None = None,
    ):
        """Print a `cappa.Exit` object to the appropriate console."""
        if e.code == 0:
            self.output(e, help=help, short_help=short_help)
        else:
            self.error(e, help=help, short_help=short_help)

    def __call__(
        self,
        message: Outputable,
        **context: Displayable | list[Displayable] | None,
    ):
        """Output a message to the `output_console`, exactly equivalent to `Output.output`."""
        self.output(message, **context)

    def output(
        self,
        message: Outputable,
        **context: Displayable | list[Displayable] | None,
    ):
        """Output a message to the `output_console`.

        Additional `**context` can be supplied into the `output_format` string template.
        """
        message = self._format_message(
            self.output_console, message, self.output_format, **context
        )
        self.write(self.output_console, message)

    def error(
        self,
        message: Outputable,
        **context: Displayable | list[Displayable] | None,
    ):
        """Output a message to the `error_console`.

        Additional `**context` can be supplied into the `error_format` string template.
        """
        message = self._format_message(
            self.error_console, message, self.error_format, **context
        )
        self.write(self.error_console, message)

    def _format_message(
        self,
        console: Console,
        message: Outputable,
        format: str,
        **context: Displayable | list[Displayable] | None,
    ) -> Text | str | None:
        code: int | str | None = None
        prog = None
        if isinstance(message, Exit):
            code = message.code
            prog = message.prog
            message = message.message

        if message is None:
            return None

        text = rich_to_ansi(console, message)

        inner_context = {
            "code": code or 0,
            "prog": prog,
            "message": text,
        }

        context = {"short_help": None, "help": None, **context}
        rendered_context = {
            k: rich_to_ansi(console, v) if v else "" for k, v in context.items()
        }
        final_context = {**inner_context, **rendered_context}

        return Text.from_markup(format.format(**final_context).strip())

    def write(self, console: Console, message: Text | str | None):
        if message is None:
            return

        console.print(message, overflow="ignore", crop=False)


def rich_to_ansi(console: Console, message: Outputable) -> str:
    with console.capture() as capture:
        if isinstance(message, list):
            for m in message:  # pyright: ignore
                console.print(m)
        else:
            console.print(message)

    return escape(capture.get().strip())



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated, TypeAliasType

import cappa
from tests.utils import Backend, backends, invoke


def foo():
    return 45


Number = TypeAliasType("Number", Annotated[int, cappa.Dep(foo)])


@dataclass
class ArgTest:
    def __call__(self, value: Number) -> int:
        return value + 1


@backends
def test_scalar(backend: Backend):
    test = invoke(ArgTest, backend=backend)
    assert test == 46



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Annotated

import cappa
from tests.utils import Backend, backends, invoke

if sys.version_info >= (3, 12):

    def foo():
        return 45

    type Number = Annotated[int, cappa.Dep(foo)]  # pyright: ignore

    @dataclass
    class ArgTest:
        def __call__(self, value: Number) -> int:
            return value + 1

    @backends
    def test_scalar(backend: Backend):
        test = invoke(ArgTest, backend=backend)
        assert test == 46



from __future__ import annotations

import contextlib
import dataclasses
import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Hashable,
    TextIO,
    cast,
    overload,
)

from rich.theme import Theme
from typing_extensions import dataclass_transform

from cappa import argparse, parser
from cappa.class_inspect import detect
from cappa.command import Alias, Command, FinalCommand
from cappa.help import HelpFormattable, HelpFormatter
from cappa.invoke.base import resolve_callable
from cappa.invoke.types import DepTypes, InvokeCallableSpec
from cappa.output import Output
from cappa.state import S, State
from cappa.types import Backend, CappaCapable, FuncOrClassDecorator, ParseResult, T, U

if TYPE_CHECKING:
    from cappa.arg import Arg, FinalArg


def create_version_arg(
    version: str | Arg[Any] | None = None,
) -> FinalArg[Any] | None:
    from dataclasses import replace

    from cappa.arg import Arg, ArgAction, Empty, Group

    if not version:
        return None

    if isinstance(version, str):
        version = Arg(
            value_name=version,
            short=["-v"],
            long=["--version"],
            help="Show the version and exit.",
            group=Group(1, "Help", section=2),
            action=ArgAction.version,
        )

    if version.value_name is Empty:
        raise ValueError(
            "Expected explicit version `Arg` to supply version number as its name, like `Arg('1.2.3', ...)`"
        )

    if version.long is True:
        version = replace(version, long="--version")

    return version.normalize(
        action=ArgAction.version, field_name="version", default=None
    )


def create_help_arg(help: bool | Arg[bool] | None = True) -> FinalArg[bool] | None:
    from cappa.arg import Arg, ArgAction, Group

    if not help:
        return None

    if isinstance(help, bool):
        help = Arg(
            short=["-h"],
            long=["--help"],
            help="Show this message and exit.",
            group=Group(0, "Help", section=2),
            action=ArgAction.help,
        )

    return help.normalize(action=ArgAction.help, field_name="help", default=None)


def create_completion_arg(
    completion: bool | Arg[bool] = True,
) -> FinalArg[bool] | None:
    from cappa.arg import Arg, ArgAction, Group

    if not completion:
        return None

    if isinstance(completion, bool):
        completion = Arg(
            long=["--completion"],
            choices=["generate", "complete"],
            group=Group(2, "Help", section=2),
            help="Use `--completion generate` to print shell-specific completion source.",
            action=ArgAction.completion,
        )

    return completion.normalize(
        field_name="completion",
        action=ArgAction.completion,
        default=None,
    )


def parse(
    obj: CappaCapable[T],
    *,
    argv: list[str] | None = None,
    input: TextIO | None = None,
    backend: Backend | None = None,
    color: bool = True,
    version: str | Arg[Any] | None = None,
    help: bool | Arg[Any] = True,
    completion: bool | Arg[Any] = True,
    theme: Theme | None = None,
    output: Output | None = None,
    help_formatter: HelpFormattable | None = None,
    state: State[Any] | None = None,
    exit_stack: contextlib.ExitStack | None = None,
) -> T:
    """Parse the command, returning an instance of `obj`.

    In the event that a subcommand is selected, only the selected subcommand
    function is invoked.

    Arguments:
        obj: A class which can represent a CLI command chain.
        argv: Defaults to the process argv. This command is generally only
            necessary when testing.
        input: Defaults to the process stdin. This command is generally only
            necessary when testing.
        backend: A function used to perform the underlying parsing and return a raw
            parsed state. This defaults to the native cappa parser, but can be changed
            to the argparse parser at `cappa.argparse.backend`.
        color: Whether to output in color.
        version: If a string is supplied, adds a -v/--version flag which returns the
            given string as the version. If an `Arg` is supplied, uses the `name`/`short`/`long`/`help`
            fields to add a corresponding version argument. Note the `name` is assumed to **be**
            the CLI's version, e.x. `Arg('1.2.3', help="Prints the version")`.
        help: If `True` (default to True), adds a -h/--help flag. If an `Arg` is supplied,
            uses the `short`/`long`/`help` fields to add a corresponding help argument.
        completion: Enables completion when using the cappa `backend` option. If `True`
            (default to True), adds a --completion flag. An `Arg` can be supplied to customize
            the argument's behavior.
        theme: Optional rich theme to customized output formatting.
        output: Optional `Output` instance. A default `Output` will constructed if one is not provided.
            Note the `color` and `theme` arguments take precedence over manually constructed `Output`
            attributes.
        help_formatter: Override the default help formatter.
        state: Optional initial State object.
        exit_stack: Optional ExitStack to use for managing context managers. If provided,
            the caller is responsible for closing the stack, allowing context to exceed the
            function call. If not provided, contexts are not entered (parse does not manage
            contexts by default).
    """
    parse_result = parse_command(
        obj=obj,
        argv=argv,
        input=input,
        backend=backend,
        color=color,
        version=version,
        help=help,
        completion=completion,
        theme=theme,
        output=output,
        help_formatter=help_formatter,
        state=state,
    )
    if exit_stack is not None:
        return exit_stack.enter_context(
            parse_result.instance.get(output=parse_result.output, managed=True)
        )
    return parse_result.instance.call(output=parse_result.output, managed=False)


async def parse_async(
    obj: CappaCapable[T],
    *,
    argv: list[str] | None = None,
    input: TextIO | None = None,
    backend: Backend | None = None,
    color: bool = True,
    version: str | Arg[Any] | None = None,
    help: bool | Arg[Any] = True,
    completion: bool | Arg[Any] = True,
    theme: Theme | None = None,
    output: Output | None = None,
    help_formatter: HelpFormattable | None = None,
    state: State[Any] | None = None,
    exit_stack: contextlib.AsyncExitStack | None = None,
) -> T:
    """Parse the command asynchronously, returning an instance of `obj`.

    This is the async version of `parse()`, necessary when using async parse functions.

    In the event that a subcommand is selected, only the selected subcommand
    function is invoked.

    Arguments:
        obj: A class which can represent a CLI command chain.
        argv: Defaults to the process argv. This command is generally only
            necessary when testing.
        input: Defaults to the process stdin. This command is generally only
            necessary when testing.
        backend: A function used to perform the underlying parsing and return a raw
            parsed state. This defaults to the native cappa parser, but can be changed
            to the argparse parser at `cappa.argparse.backend`.
        color: Whether to output in color.
        version: If a string is supplied, adds a -v/--version flag which returns the
            given string as the version. If an `Arg` is supplied, uses the `name`/`short`/`long`/`help`
            fields to add a corresponding version argument. Note the `name` is assumed to **be**
            the CLI's version, e.x. `Arg('1.2.3', help="Prints the version")`.
        help: If `True` (default to True), adds a -h/--help flag. If an `Arg` is supplied,
            uses the `short`/`long`/`help` fields to add a corresponding help argument.
        completion: Enables completion when using the cappa `backend` option. If `True`
            (default to True), adds a --completion flag. An `Arg` can be supplied to customize
            the argument's behavior.
        theme: Optional rich theme to customized output formatting.
        output: Optional `Output` instance. A default `Output` will constructed if one is not provided.
            Note the `color` and `theme` arguments take precedence over manually constructed `Output`
            attributes.
        help_formatter: Override the default help formatter.
        state: Optional initial State object.
        exit_stack: Optional AsyncExitStack to use for managing async context managers.
            If provided, the caller is responsible for closing the stack, allowing context
            to exceed the function call. If not provided, contexts are not entered (parse does
            not manage contexts by default).
    """
    parse_result = parse_command(
        obj=obj,
        argv=argv,
        input=input,
        backend=backend,
        color=color,
        version=version,
        help=help,
        completion=completion,
        theme=theme,
        output=output,
        help_formatter=help_formatter,
        state=state,
    )
    if exit_stack is not None:
        return await exit_stack.enter_async_context(
            parse_result.instance.get_async(output=parse_result.output, managed=True)
        )
    return await parse_result.instance.call_async(
        output=parse_result.output, managed=False
    )


def invoke(
    obj: CappaCapable[T],
    *,
    deps: DepTypes = None,
    argv: list[str] | None = None,
    input: TextIO | None = None,
    backend: Backend | None = None,
    color: bool = True,
    version: str | Arg[Any] | None = None,
    help: bool | Arg[Any] = True,
    completion: bool | Arg[Any] = True,
    theme: Theme | None = None,
    output: Output | None = None,
    help_formatter: HelpFormattable | None = None,
    state: State[Any] | None = None,
    exit_stack: contextlib.ExitStack | None = None,
) -> Any:
    """Parse the command, and invoke the selected async command or subcommand.

    In the event that a subcommand is selected, only the selected subcommand
    function is invoked.

    Arguments:
        obj: A class which can represent a CLI command chain.
        deps: Optional extra depdnencies to load ahead of invoke processing. These
            deps are evaluated in order and unconditionally.
        argv: Defaults to the process argv. This command is generally only
            necessary when testing.
        input: Defaults to the process stdin. This command is generally only
            necessary when testing.
        backend: A function used to perform the underlying parsing and return a raw
            parsed state. This defaults to the native cappa parser, but can be changed
            to the argparse parser at `cappa.argparse.backend`.
        color: Whether to output in color.
        version: If a string is supplied, adds a -v/--version flag which returns the
            given string as the version. If an `Arg` is supplied, uses the `name`/`short`/`long`/`help`
            fields to add a corresponding version argument. Note the `name` is assumed to **be**
            the CLI's version, e.x. `Arg('1.2.3', help="Prints the version")`.
        help: If `True` (default to True), adds a -h/--help flag. If an `Arg` is supplied,
            uses the `short`/`long`/`help` fields to add a corresponding help argument.
        completion: Enables completion when using the cappa `backend` option. If `True`
            (default to True), adds a --completion flag. An `Arg` can be supplied to customize
            the argument's behavior.
        theme: Optional rich theme to customized output formatting.
        output: Optional `Output` instance. A default `Output` will constructed if one is not provided.
            Note the `color` and `theme` arguments take precedence over manually constructed `Output`
            attributes.
        help_formatter: Override the default help formatter.
        state: Optional initial State object.
        exit_stack: Optional ExitStack to use for managing context managers. If provided,
            the caller is responsible for closing the stack, allowing context to exceed the
            function call. If not provided, a new stack is created and automatically closed.
    """
    parse_result = parse_command(
        obj=obj,
        argv=argv,
        input=input,
        backend=backend,
        color=color,
        version=version,
        help=help,
        completion=completion,
        theme=theme,
        output=output,
        help_formatter=help_formatter,
        state=state,
    )

    def _invoke_with_stack(stack: contextlib.ExitStack):
        instance = stack.enter_context(
            parse_result.instance.get(output=parse_result.output)
        )

        # Resolve all implicit deps
        resolved_implicit_deps: dict[Hashable, Any] = {}
        for key, resolved_dep in parse_result.implicit_deps.items():
            resolved_implicit_deps[key] = stack.enter_context(
                resolved_dep.get(output=parse_result.output)
            )

        resolved, global_deps = resolve_callable(
            parse_result.root_command,
            parse_result.parsed_command,
            instance,
            implicit_deps=resolved_implicit_deps,
            output=parse_result.output,
            state=parse_result.state,
            deps=deps,
        )
        for dep in global_deps:
            stack.enter_context(dep.get(output=parse_result.output))

        return stack.enter_context(resolved.get(output=parse_result.output))

    if exit_stack is not None:
        return _invoke_with_stack(exit_stack)

    with contextlib.ExitStack() as stack:
        return _invoke_with_stack(stack)


async def invoke_async(
    obj: CappaCapable[T],
    *,
    deps: DepTypes = None,
    argv: list[str] | None = None,
    input: TextIO | None = None,
    backend: Backend | None = None,
    color: bool = True,
    version: str | Arg[str] | None = None,
    help: bool | Arg[bool] = True,
    completion: bool | Arg[bool] = True,
    theme: Theme | None = None,
    output: Output | None = None,
    help_formatter: HelpFormattable | None = None,
    state: State[Any] | None = None,
    exit_stack: contextlib.AsyncExitStack | None = None,
) -> Any:
    """Parse the command, and invoke the selected command or subcommand.

    In the event that a subcommand is selected, only the selected subcommand
    function is invoked.

    Arguments:
        obj: A class which can represent a CLI command chain.
        deps: Optional extra depdnencies to load ahead of invoke processing. These
            deps are evaluated in order and unconditionally.
        argv: Defaults to the process argv. This command is generally only
            necessary when testing.
        input: Defaults to the process stdin. This command is generally only
            necessary when testing.
        backend: A function used to perform the underlying parsing and return a raw
            parsed state. This defaults to the native cappa parser, but can be changed
            to the argparse parser at `cappa.argparse.backend`.
        color: Whether to output in color.
        version: If a string is supplied, adds a -v/--version flag which returns the
            given string as the version. If an `Arg` is supplied, uses the `name`/`short`/`long`/`help`
            fields to add a corresponding version argument. Note the `name` is assumed to **be**
            the CLI's version, e.x. `Arg('1.2.3', help="Prints the version")`.
        help: If `True` (default to True), adds a -h/--help flag. If an `Arg` is supplied,
            uses the `short`/`long`/`help` fields to add a corresponding help argument.
        completion: Enables completion when using the cappa `backend` option. If `True`
            (default to True), adds a --completion flag. An `Arg` can be supplied to customize
            the argument's behavior.
        theme: Optional rich theme to customized output formatting.
        output: Optional `Output` instance. A default `Output` will constructed if one is not provided.
            Note the `color` and `theme` arguments take precedence over manually constructed `Output`
            attributes.
        help_formatter: Override the default help formatter.
        state: Optional initial State object.
        exit_stack: Optional AsyncExitStack to use for managing async context managers.
            If provided, the caller is responsible for closing the stack, allowing context to
            exceed the function call. If not provided, a new stack is created and automatically closed.
    """
    parse_result = parse_command(
        obj=obj,
        argv=argv,
        input=input,
        backend=backend,
        color=color,
        version=version,
        help=help,
        completion=completion,
        theme=theme,
        output=output,
        help_formatter=help_formatter,
        state=state,
    )

    async def _invoke_async_with_stack(stack: contextlib.AsyncExitStack):
        instance = await stack.enter_async_context(
            parse_result.instance.get_async(output=parse_result.output)
        )

        # Resolve all implicit deps
        resolved_implicit_deps: dict[Hashable, Any] = {}
        for key, resolved_dep in parse_result.implicit_deps.items():
            resolved_implicit_deps[key] = await stack.enter_async_context(
                resolved_dep.get_async(output=parse_result.output)
            )

        resolved, global_deps = resolve_callable(
            parse_result.root_command,
            parse_result.parsed_command,
            instance,
            implicit_deps=resolved_implicit_deps,
            output=parse_result.output,
            state=parse_result.state,
            deps=deps,
        )
        for dep in global_deps:
            await stack.enter_async_context(dep.get_async(output=parse_result.output))

        return await stack.enter_async_context(
            resolved.get_async(output=parse_result.output)
        )

    if exit_stack is not None:
        return await _invoke_async_with_stack(exit_stack)

    async with contextlib.AsyncExitStack() as stack:
        return await _invoke_async_with_stack(stack)


def parse_command(
    obj: CappaCapable[T],
    *,
    argv: list[str] | None = None,
    input: TextIO | None = None,
    backend: Backend | None = None,
    color: bool = True,
    version: str | Arg[str] | None = None,
    help: bool | Arg[bool] = True,
    completion: bool | Arg[bool] = True,
    theme: Theme | None = None,
    output: Output | None = None,
    help_formatter: HelpFormattable | None = None,
    state: State[S] | None = None,
) -> ParseResult[T, S]:
    concrete_backend = _coalesce_backend(backend)
    concrete_output = _coalesce_output(output, theme, color)
    concrete_state: State[S] = State.ensure(state)  # type: ignore

    command: FinalCommand[T] = collect(
        obj,
        help=help,
        version=version,
        completion=completion,
        backend=concrete_backend,
        help_formatter=help_formatter,
        state=concrete_state,
    )
    return command.parse_command(
        argv=argv,
        input=input,
        backend=concrete_backend,
        output=concrete_output,
        state=concrete_state,
    )


@overload
def command(
    _cls: type[T],
    *,
    name: str | None = None,
    aliases: list[str | Alias] | None = None,
    help: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    invoke: InvokeCallableSpec[Any] | None = None,
    hidden: bool = False,
    default_short: bool = False,
    default_long: bool = False,
    deprecated: bool = False,
    help_formatter: HelpFormattable = HelpFormatter.default,
) -> type[T]: ...
@overload
def command(
    *,
    name: str | None = None,
    aliases: list[str | Alias] | None = None,
    help: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    invoke: InvokeCallableSpec[Any] | None = None,
    hidden: bool = False,
    default_short: bool = False,
    default_long: bool = False,
    deprecated: bool = False,
    help_formatter: HelpFormattable = HelpFormatter.default,
) -> FuncOrClassDecorator: ...
@overload
def command(
    _cls: T,
    *,
    name: str | None = None,
    aliases: list[str | Alias] | None = None,
    help: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    invoke: InvokeCallableSpec[Any] | None = None,
    hidden: bool = False,
    default_short: bool = False,
    default_long: bool = False,
    deprecated: bool = False,
    help_formatter: HelpFormattable = HelpFormatter.default,
) -> T: ...


@dataclass_transform()  # type: ignore[misc]
def command(
    _cls: type[T] | T | None = None,
    *,
    name: str | None = None,
    aliases: list[str | Alias] | None = None,
    help: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    invoke: InvokeCallableSpec[Any] | None = None,
    hidden: bool = False,
    default_short: bool = False,
    default_long: bool = False,
    deprecated: bool = False,
    help_formatter: HelpFormattable = HelpFormatter.default,
) -> type[T] | T | FuncOrClassDecorator:
    """Register a cappa CLI command/subcomment.

    Args:
        name: The name of the command. If omitted, the name of the command
            will be the name of the `cls`, converted to dash-case.
        aliases: Alternate names that may be used to invoke this command as a
            subcommand. Each entry is either a plain string (visible alias) or
            an :class:`Alias` instance (which can be marked `hidden=True` or
            `deprecated=...`). Only meaningful when the class is used as a
            subcommand.
        help: Optional help text. If omitted, the `cls` docstring will be parsed,
            and the headline section will be used to document the command
            itself, and the arguments section will become the default help text for
            any params/options.
        description: Optional extended help text. If omitted, the `cls` docstring will
            be parsed, and the extended long description section will be used.
        epilog: Optional text displayed after the argument list in the help output.
        invoke: Optional command to be called in the event parsing is successful.
            In the case of subcommands, it will only call the parsed/selected
            function to invoke.
        hidden: If `True`, the command will not be included in the help output.
            This option is only relevant to subcommands.
        default_short: If `True`, all arguments will be treated as though annotated
            with `Annotated[T, Arg(short=True)]`, unless otherwise annotated.
        default_long: If `True`, all arguments will be treated as though annotated
            with `Annotated[T, Arg(long=True)]`, unless otherwise annotated.
        deprecated: If supplied, the argument will be marked as deprecated. If given `True`,
            a default message will be generated, otherwise a supplied string will be
            used as the deprecation message.
        help_formatter: Override the default help formatter.
    """

    def wrapper(_decorated_cls: U) -> U:
        if inspect.isclass(_decorated_cls) and not detect(_decorated_cls):
            _decorated_cls = dataclasses.dataclass(_decorated_cls)  # type: ignore

        command: Command[T] = Command.get(_decorated_cls)  # type: ignore
        instance = dataclasses.replace(
            command,
            invoke=invoke,
            name=name,
            aliases=list(aliases) if aliases is not None else command.aliases,
            help=help,
            description=description,
            epilog=epilog,
            hidden=hidden,
            default_short=default_short,
            default_long=default_long,
            deprecated=deprecated,
            help_formatter=help_formatter,
        )
        _decorated_cls.__cappa__ = instance  # type: ignore

        # Functions (and in particular class methods, must return a function object in order
        # to be attached as methods) cannot be nested, so we can just directly return it.
        if inspect.isfunction(_decorated_cls):
            return cast(U, _decorated_cls)

        # Whereas classes will **generally** be the **exact** object as `_decorated_cls` was,
        # except in the case of dynamically generated subclasses used for detecting methods.
        return instance.cmd_cls  # type: ignore

    if _cls is not None:
        return wrapper(_cls)
    return wrapper


def collect(
    obj: CappaCapable[T],
    *,
    backend: Backend | None = None,
    version: str | Arg[str] | None = None,
    help: bool | Arg[bool] = True,
    completion: bool | Arg[bool] = True,
    help_formatter: HelpFormattable | None = None,
    state: State[Any] | None = None,
) -> FinalCommand[T]:
    """Retrieve the `Command` object from a cappa-capable source class.

    Arguments:
        obj: A class which can represent a CLI command chain.
        backend: A function used to perform the underlying parsing and return a raw
            parsed state. This defaults to constructing built-in function using argparse.
        version: If a string is supplied, adds a -v/--version flag which returns the
            given string as the version. If an `Arg` is supplied, uses the `name`/`short`/`long`/`help`
            fields to add a corresponding version argument. Note the `name` is assumed to **be**
            the CLI's version, e.x. `Arg('1.2.3', help="Prints the version")`.
        help: If `True` (default to True), adds a -h/--help flag. If an `Arg` is supplied,
            uses the `short`/`long`/`help` fields to add a corresponding help argument.
        completion: Enables completion when using the cappa `backend` option. If `True`
            (default to True), adds a --completion flag. An `Arg` can be supplied to customize
            the argument's behavior.
        color: Whether to output in color.
        help_formatter: Override the default help formatter.
        state: Optional initial State object.
    """
    state = State.ensure(state)  # pyright: ignore

    command: FinalCommand[T] = Command.get(  # pyright: ignore
        obj, help_formatter=help_formatter
    ).collect(state=state)

    concrete_backend = _coalesce_backend(backend)
    if concrete_backend is argparse.backend:  # pyright: ignore
        completion = False

    help_arg = create_help_arg(help)
    version_arg = create_version_arg(version)
    completion_arg = create_completion_arg(completion)

    return command.add_meta_actions(
        help=help_arg, version=version_arg, completion=completion_arg
    )


def _coalesce_backend(backend: Backend | None = None) -> Backend:
    if backend is None:  # pragma: no cover
        return parser.backend
    return backend


def _coalesce_output(
    output: Output | None = None, theme: Theme | None = None, color: bool = True
):
    if output is None:
        output = Output()

    output.theme(theme)

    if not color:
        output.color(False)

    return output



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


def test_unbounded_list_option():
    @dataclass
    class Args:
        unbounded: Annotated[list[list[int]], cappa.Arg(short=True)]

    result = parse(Args, "-u", "0")
    assert result == Args([[0]])

    result = parse(Args, "-u", "0", "1", "2", "3", "4", "5")
    assert result == Args([[0, 1, 2, 3, 4, 5]])


def test_unbounded_set_option():
    @dataclass
    class Args:
        unbounded: Annotated[set[tuple[int, int]], cappa.Arg(short=True)]

    result = parse(Args, "-u", "0", "1")
    assert result == Args({(0, 1)})


def test_unbounded_tuple_option():
    @dataclass
    class Args:
        unbounded: Annotated[tuple[list[int], ...], cappa.Arg(short=True)]

    result = parse(Args, "-u", "0")
    assert result == Args(([0],))

    result = parse(Args, "-u", "0", "1", "2", "3", "4", "5")
    assert result == Args(([0, 1, 2, 3, 4, 5],))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import cappa
from tests.utils import Backend, backends, invoke


def foo(command: Command):
    print("called", command)


def command():
    return 1


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_global_dep_string_reference(capsys: Any, backend: Backend):
    result = invoke(
        Command,
        deps=["tests.invoke.test_global_dep_string_reference.foo"],
        backend=backend,
    )
    assert result == 1

    out = capsys.readouterr().out
    assert "called Command()" in out



from __future__ import annotations

import io
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, BinaryIO, TextIO
from unittest.mock import mock_open, patch

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@contextmanager
def file_content(content: str, mode: str = "r"):
    if "b" in mode:
        content = content.encode("utf-8")  # type: ignore

    mock = mock_open(read_data=content)
    with patch("builtins.open", new=mock):
        yield mock

    for call in mock.call_args[::2]:
        assert call[1] == mode


@contextmanager
def stdin(content: str):
    with patch("sys.stdin", new=io.TextIOWrapper(io.BytesIO(content.encode("utf-8")))):
        yield


@backends
def test_text_io_default(backend: Backend):
    @dataclass
    class Foo:
        bar: TextIO

    with file_content("wat"):
        test = parse(Foo, "foo.py", backend=backend)

    assert test.bar.read() == "wat"


@backends
def test_text_io(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[TextIO, cappa.FileMode(mode="r")]

    with file_content("wat"):
        test = parse(Foo, "foo.py", backend=backend)

    assert test.bar.read() == "wat"


@backends
def test_text_io_write(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="w")]

    with file_content("wat", mode="w"):
        test = parse(Foo, "foo.py", backend=backend)

    test.bar.write(b"wat")


@backends
def test_binary_io_default(backend: Backend):
    @dataclass
    class Foo:
        bar: BinaryIO

    with file_content("wat", mode="rb"):
        test = parse(Foo, "foo.py", backend=backend)

    assert test.bar.read() == b"wat"


@backends
def test_binary_io(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="rb")]

    with file_content("wat", mode="rb"):
        test = parse(Foo, "foo.py", backend=backend)

    assert test.bar.read() == b"wat"


@backends
def test_binary_io_write(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="wb")]

    with file_content("wat", mode="wb"):
        test = parse(Foo, "foo.py", backend=backend)

    test.bar.write(b"wat")


@backends
def test_stdin(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="r")]

    with stdin("wat"):
        test = parse(Foo, "-", backend=backend)
        assert test.bar.read() == "wat"


@backends
def test_stdin_binary(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="rb")]

    with stdin("wat"):
        test = parse(Foo, "-", backend=backend)
        assert test.bar.read() == b"wat"


@backends
def test_stdout(backend: Backend, capsys: Any):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="w")]

    test = parse(Foo, "-", backend=backend)
    test.bar.write("wat")  # type: ignore

    out = capsys.readouterr().out
    assert out == "wat"


@backends
def test_stdout_binary(backend: Backend, capsys: Any):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="wb")]

    test = parse(Foo, "-", backend=backend)
    test.bar.write(b"wat")

    out = capsys.readouterr().out
    assert out == "wat"


@backends
def test_invalid_mode_dash(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="tb")]

    with pytest.raises(cappa.Exit) as e:
        parse(Foo, "-", backend=backend)

    assert e.value.message == "Invalid mode 'tb' with supplied '-' file name."


@backends
def test_invalid_mode(backend: Backend):
    @dataclass
    class Foo:
        bar: Annotated[BinaryIO, cappa.FileMode(mode="tb")]

    with pytest.raises(cappa.Exit) as e:
        parse(Foo, "foo.py", backend=backend)

    assert str(e.value.message).startswith("Invalid value for 'bar'")
    assert str(e.value.message).endswith("can't have text and binary mode at once")


@backends
def test_open_oserror(backend: Backend):
    @dataclass
    class Foo:
        bar: BinaryIO

    with pytest.raises(cappa.Exit) as e:
        parse(Foo, "thisshouldneverexist.py", backend=backend)

    assert (
        e.value.message
        == "Cannot open thisshouldneverexist.py: [Errno 2] No such file or directory: 'thisshouldneverexist.py'"
    )



from __future__ import annotations

import contextlib
import dataclasses
import sys
from collections.abc import Callable
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generator,
    Generic,
    Hashable,
    Iterable,
    Protocol,
    Sequence,
    TextIO,
    TypedDict,
    TypeVar,
    cast,
)

from type_lens.type_view import TypeView

from cappa.arg import Arg, FinalArg, Group
from cappa.class_inspect import fields as get_fields
from cappa.class_inspect import get_command, get_command_capable_object
from cappa.docstring import ClassHelpText
from cappa.help import HelpFormattable, HelpFormatter
from cappa.invoke.types import Resolved
from cappa.output import Exit, Output
from cappa.state import S, State
from cappa.subcommand import FinalSubcommand, Subcommand
from cappa.type_view import CallableView
from cappa.types import ParseResult
from cappa.typing import assert_type

if TYPE_CHECKING:
    from cappa.base import Backend, CappaCapable

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class Alias:
    """Describe an alternate name for a subcommand.

    Arguments:
        name: The alternate name a user may type to invoke the subcommand.
        hidden: If `True`, the alias is accepted but omitted from help output and
            from shell completion. Useful for supporting old names without
            advertising them.
        deprecated: If supplied, invoking the subcommand via this alias will emit
            a deprecation warning. If `True`, a default message is used; a string
            value is used as the deprecation message verbatim.
    """

    name: str
    hidden: bool = False
    deprecated: bool | str = False

    @classmethod
    def coerce(cls, value: str | Alias) -> Alias:
        if isinstance(value, Alias):
            return value
        return cls(name=value)


class CommandArgs(TypedDict, total=False):
    cmd_cls: type
    arguments: list[Arg[Any] | Subcommand]
    name: str | None
    aliases: list[str | Alias]
    help: str | None
    description: str | None
    invoke: Callable[..., Any] | str | None

    hidden: bool
    default_short: bool
    default_long: bool


@dataclasses.dataclass
class Command(Generic[T]):
    """Register a cappa CLI command/subcomment.

    Args:
        cmd_cls: The class representing the command/subcommand
        name: The name of the command. If omitted, the name of the command
            will be the name of the `cls`, converted to dash-case.
        help: Optional help text. If omitted, the `cls` docstring will be parsed,
            and the headline section will be used to document the command
            itself, and the arguments section will become the default help text for
            any params/options.
        description: Optional extended help text. If omitted, the `cls` docstring will
            be parsed, and the extended long description section will be used.
        epilog: Optional text displayed after the argument list in the help output.
        invoke: Optional command to be called in the event parsing is successful.
            In the case of subcommands, it will only call the parsed/selected
            function to invoke. The value can **either** be a callable object or
            a string. When the value is a string it will be interpreted as
            `module.submodule.function`; the module will be dynamically imported,
            and the referenced function invoked.
        hidden: If `True`, the command will not be included in the help output.
            This option is only relevant to subcommands.
        default_short: If `True`, all arguments will be treated as though annotated
            with `Annotated[T, Arg(short=True)]`, unless otherwise annotated.
        default_long: If `True`, all arguments will be treated as though annotated
            with `Annotated[T, Arg(long=True)]`, unless otherwise annotated.
        deprecated: If supplied, the argument will be marked as deprecated. If given `True`,
            a default message will be generated, otherwise a supplied string will be
            used as the deprecation message.
    """

    cmd_cls: type[T]
    arguments: Sequence[Arg[Any] | Subcommand | FinalDestructure[Any]] = (
        dataclasses.field(default_factory=lambda: [])
    )
    propagated_arguments: list[FinalArg[Any]] = dataclasses.field(
        default_factory=lambda: []
    )

    name: str | None = None
    aliases: list[str | Alias] = dataclasses.field(default_factory=lambda: [])
    help: str | None = None
    description: str | None = None
    epilog: str | None = None
    invoke: Callable[..., Any] | str | None = None

    hidden: bool = False
    default_short: bool = False
    default_long: bool = False
    deprecated: bool | str = False

    help_formatter: HelpFormattable = HelpFormatter.default

    _collected: bool = False

    @classmethod
    def get(
        cls, obj: CappaCapable[T], help_formatter: HelpFormattable | None = None
    ) -> Command[T]:
        help_formatter = help_formatter or HelpFormatter.default

        instance = None
        if isinstance(obj, cls):
            instance = obj
        else:
            obj = get_command_capable_object(obj)
            instance = get_command(obj)

        if instance:
            return dataclasses.replace(instance, help_formatter=help_formatter)

        assert not isinstance(obj, Command)
        return cls(
            obj,  # pyright: ignore
            help_formatter=help_formatter,
        )

    def real_name(self) -> str:
        if self.name is not None:
            return self.name

        import re

        cls_name = self.cmd_cls.__name__
        return re.sub(r"(?<!^)(?=[A-Z])", "-", cls_name).lower()

    def resolved_aliases(self) -> list[Alias]:
        return [Alias.coerce(a) for a in self.aliases]

    def collect(
        self,
        propagated_arguments: list[FinalArg[Any]] | None = None,
        state: State[Any] | None = None,
    ) -> FinalCommand[T]:
        kwargs: CommandArgs = CommandArgs()

        help_text = ClassHelpText.collect(self.cmd_cls)

        if not self.help:
            kwargs["help"] = help_text.summary

        if not self.description:
            kwargs["description"] = help_text.body

        fields = get_fields(self.cmd_cls)
        function_view = CallableView.from_callable(self.cmd_cls, include_extras=True)

        propagated_arguments = propagated_arguments or []

        arguments: list[FinalArg[Any] | FinalDestructure[Any]] = []
        raw_subcommands: list[tuple[Subcommand, TypeView[Any] | None, str | None]] = []
        if self.arguments:
            param_by_name = {p.name: p for p in function_view.parameters}
            for arg in self.arguments:
                arg_help = help_text.args.get(assert_type(arg.field_name, str))
                if isinstance(arg, Arg):
                    type_view = (
                        param_by_name[cast(str, arg.field_name)].type_view
                        if arg.field_name in param_by_name
                        else None
                    )
                    arguments.append(
                        arg.normalize(
                            type_view=type_view,
                            default_short=self.default_short,
                            default_long=self.default_long,
                            fallback_help=arg_help,
                            state=state,
                        )
                    )
                elif isinstance(arg, FinalDestructure):
                    arguments.append(arg)
                else:
                    raw_subcommands.append((arg, None, None))

        else:
            param_by_name = {p.name: p for p in function_view.parameters}
            for field in fields:
                param_view = param_by_name[field.name]
                arg_help = help_text.args.get(param_view.name)

                maybe_subcommand = Subcommand.detect(
                    field,
                    param_view.type_view,
                )
                if maybe_subcommand:
                    raw_subcommands.append(
                        (
                            maybe_subcommand,
                            param_view.type_view,
                            field.name,
                        )
                    )
                else:
                    arg_defs = Arg.collect(
                        field,
                        param_view.type_view,
                        fallback_help=arg_help,
                        default_short=self.default_short,
                        default_long=self.default_long,
                        state=state,
                    )
                    arguments.extend(arg_defs)

        propagating_arguments = [
            *propagated_arguments,
            *(arg for arg in arguments if isinstance(arg, FinalArg) and arg.propagate),
        ]
        subcommands = [
            subcommand.normalize(
                type_view,
                field_name,
                help_formatter=self.help_formatter,
                propagated_arguments=propagating_arguments,
                state=state,
            )
            for subcommand, type_view, field_name in raw_subcommands
        ]

        check_group_identity([a for a in arguments if isinstance(a, FinalArg)])
        final_arguments: list[
            FinalArg[Any] | FinalSubcommand | FinalDestructure[Any]
        ] = [
            *arguments,
            *subcommands,
        ]

        return FinalCommand(
            cmd_cls=self.cmd_cls,
            arguments=final_arguments,
            propagated_arguments=propagated_arguments,
            name=self.name,
            aliases=self.aliases,
            help=kwargs.get("help", self.help),
            description=kwargs.get("description", self.description),
            epilog=self.epilog,
            invoke=self.invoke,
            hidden=self.hidden,
            default_short=self.default_short,
            default_long=self.default_long,
            deprecated=self.deprecated,
            help_formatter=self.help_formatter,
            _collected=self._collected,
        )


@dataclasses.dataclass
class FinalCommand(Command[T]):
    """Post-normalization form of :class:`Command` with narrowed field types.

    Produced exclusively by :meth:`Command.collect`.
    """

    arguments: Sequence[FinalArg[Any] | FinalSubcommand | FinalDestructure[Any]] = (  # pyright: ignore
        dataclasses.field(default_factory=list)
    )
    propagated_arguments: list[FinalArg[Any]] = dataclasses.field(  # pyright: ignore
        default_factory=list
    )

    @property
    def subcommand(self) -> FinalSubcommand | None:
        return next(
            (arg for arg in self.arguments if isinstance(arg, FinalSubcommand)),
            None,
        )

    @property
    def value_arguments(self) -> Iterable[FinalArg[Any]]:
        for arg in self.arguments:
            if isinstance(arg, FinalArg) and arg.has_value:
                yield arg

    @property
    def destructured_arguments(self) -> Iterable[FinalDestructure[Any]]:
        for arg in self.arguments:
            if isinstance(arg, FinalDestructure):
                yield arg

    @property
    def all_arguments(self) -> Iterable[FinalArg[Any] | FinalSubcommand]:
        for arg in self.arguments:
            if not isinstance(arg, FinalDestructure):
                yield arg
        for arg in self.propagated_arguments:
            yield arg

    @property
    def options(self) -> Iterable[FinalArg[Any]]:
        for arg in self.arguments:
            if isinstance(arg, FinalArg) and arg.is_option:
                yield arg

    @property
    def positional_arguments(
        self,
    ) -> Iterable[FinalArg[Any] | FinalSubcommand]:
        for arg in self.arguments:
            if (
                isinstance(arg, FinalArg) and not arg.short and not arg.long
            ) or isinstance(arg, FinalSubcommand):
                yield arg

    def add_meta_actions(
        self,
        help: FinalArg[bool] | None = None,
        version: FinalArg[str] | None = None,
        completion: FinalArg[bool] | None = None,
    ):
        if self._collected:
            return self

        arguments = [
            dataclasses.replace(
                arg,
                options={
                    name: option.add_meta_actions(help)
                    for name, option in arg.options.items()
                },
            )
            if help and isinstance(arg, FinalSubcommand)
            else arg
            for arg in self.arguments
        ]

        if help:
            arguments.append(help)
        if version:
            arguments.append(version)
        if completion:
            arguments.append(completion)
        return dataclasses.replace(self, arguments=arguments, _collected=True)

    def map_result(
        self,
        command: FinalCommand[T],
        prog: str,
        parsed_args: dict[str, Any],
        output: Output,
        state: State[Any] | None = None,
        input: TextIO | None = None,
    ) -> tuple[Resolved[T], dict[Hashable, Any]]:
        state = State.ensure(state)  # pyright: ignore

        kwargs: dict[str, Any] = {}
        for arg in self.value_arguments:
            kwargs[arg.field_name] = arg.map_result(
                prog, parsed_args, state=state, input=input
            )

        subcommand_deps: dict[Hashable, Any] = {}
        for destructure in self.destructured_arguments:
            fd_parsed = parsed_args.get(destructure.field_name, {})
            fd_resolved = destructure.map_result(
                prog, fd_parsed, output, state=state, input=input
            )
            kwargs[destructure.field_name] = fd_resolved

        subcommand = self.subcommand
        if subcommand:
            field_name = subcommand.field_name
            if field_name in parsed_args:
                value = parsed_args[field_name]
                value, subcommand_deps = subcommand.map_result(
                    prog, value, output=output, state=state
                )
                kwargs[field_name] = value

        def map_result(**kwargs: dict[str, Any]) -> T:
            with graceful_exit(command, prog, output):
                return command.cmd_cls(**kwargs)

        resolved = Resolved(map_result, kwargs=kwargs)
        key = cast(Hashable, command.cmd_cls)
        deps: dict[Hashable, Any] = {key: resolved, **subcommand_deps}
        return resolved, deps

    def parse_command(
        self,
        *,
        output: Output,
        backend: Backend,
        argv: list[str] | None = None,
        input: TextIO | None = None,
        state: State[S] | None = None,
    ) -> ParseResult[T, S]:
        if argv is None:  # pragma: no cover
            argv = sys.argv[1:]

        prog = self.real_name()
        result_state: State[S] = State.ensure(state)  # type: ignore

        with graceful_exit(self, prog, output):
            parser, parsed_command, parsed_args = backend(
                self, argv, output=output, prog=prog
            )
            prog = parser.prog
            result, implicit_deps = self.map_result(
                self, prog, parsed_args, state=state, input=input, output=output
            )

        return ParseResult(
            root_command=self,
            parsed_command=parsed_command,
            instance=result,
            implicit_deps=implicit_deps,
            output=output,
            state=result_state,
        )


H = TypeVar("H", covariant=True)


class HasCommand(Generic[H], Protocol):
    __cappa__: ClassVar[Command[Any]]


def check_group_identity(args: list[FinalArg[Any]]):
    group_identity: dict[str, Group] = {}

    for arg in args:
        identity = group_identity.get(arg.group.id)
        if identity and identity != arg.group:
            raise ValueError(
                f"Group details do not match among arguments: {arg.group.diff(identity)}."
            )

        group_identity[arg.group.id] = arg.group


@contextlib.contextmanager
def graceful_exit(
    command: FinalCommand[T], prog: str, output: Output
) -> Generator[None, None, None]:
    try:
        yield
    except BaseException as e:
        if isinstance(e, Exit):
            command = e.command or command
            prog = e.prog or prog

        help = command.help_formatter.long(command, prog)
        short_help = command.help_formatter.short(command, prog)

        if isinstance(e, ValueError):
            exc = Exit(str(e), code=2, prog=prog, command=command)
            output.exit(exc, help=help, short_help=short_help)
            raise exc

        if isinstance(e, Exit):
            output.exit(e, help=help, short_help=short_help)
            raise

        raise


from cappa.destructure import FinalDestructure  # noqa: E402



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_manually_specified_choices(backend: Backend):
    @dataclass
    class ArgTest:
        choice: Annotated[str, cappa.Arg(choices=["a", "1"])]

    result = parse(ArgTest, "a")
    assert result.choice == "a"

    result = parse(ArgTest, "1")
    assert result.choice == "1"

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "two", backend=backend)

    message = str(e.value.message).lower()
    assert "invalid choice: 'two' (choose from 'a', '1')" in message


@backends
def test_optional_set_of_choices(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        choice: Annotated[set[Literal["one", "two"]] | None, cappa.Arg(short=True)] = (
            None
        )

    with pytest.raises(cappa.HelpExit):
        parse(ArgTest, "--help", backend=backend)

    result = capsys.readouterr().out
    assert "Valid options: one, two." in result


@backends
def test_variadic_tuple_choices(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        choice: Annotated[
            tuple[Literal["one", "two"], ...] | None, cappa.Arg(short=True)
        ] = None

    with pytest.raises(cappa.HelpExit):
        parse(ArgTest, "--help", backend=backend)

    result = capsys.readouterr().out
    assert "Valid options: one, two." in result


@backends
def test_tuple_choices(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        choice: Annotated[
            tuple[Literal["one", "two"], int] | None, cappa.Arg(short=True)
        ] = None

    with pytest.raises(cappa.HelpExit):
        parse(ArgTest, "--help", backend=backend)

    result = capsys.readouterr().out
    assert "Valid options: one, two." not in result


@backends
def test_literal_parse_sequence(backend: Backend):
    @dataclass
    class LiteralParse:
        log_level: Annotated[
            str,
            cappa.Arg(
                short="-L", long=True, choices=["DEBUG"], parse=[str.upper, str.strip]
            ),
        ] = "DEBUG"

    result = parse(LiteralParse, "--log-level", "  debug  ", backend=backend)
    assert result == LiteralParse(log_level="DEBUG")


explicit_choice_default = ["1s", "1m"]


@backends
def test_explicit_choice_sequence(backend: Backend):
    @dataclass
    class Example:
        interval: Annotated[
            list[str] | None, cappa.Arg(choices=explicit_choice_default, default=["1m"])
        ]

    result = parse(Example, backend=backend)
    assert result == Example(interval=["1m"])

    result = parse(Example, "1s", backend=backend)
    assert result == Example(interval=["1s"])

    with pytest.raises(cappa.Exit) as e:
        parse(Example, "2s", backend=backend)

    assert (
        e.value.message
        == "Invalid value for 'interval': Invalid choice: '2s' (choose from '1s', '1m')"
    )



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Config:
    verbose: Annotated[int, cappa.Arg(short=True, count=True)]


@dataclass
class Count:
    config: cappa.Destructured[Config]


def test_count():
    test = parse(Count, "-v")
    assert test == Count(config=Config(verbose=1))

    test = parse(Count, "-vvvvv")
    assert test == Count(config=Config(verbose=5))



from __future__ import annotations

import dataclasses
import functools
import inspect
import sys
import typing
from enum import Enum
from typing import TYPE_CHECKING, Any

from typing_extensions import Annotated, Self

from cappa.output import Output
from cappa.state import State
from cappa.type_view import CallableView, Empty, EmptyType
from cappa.typing import T, find_annotations

if typing.TYPE_CHECKING:
    from cappa.command import Command

__all__ = [
    "detect",
    "fields",
]


def detect(cls: type) -> bool:
    return bool(ClassTypes.from_cls(cls))


@dataclasses.dataclass
class Field:
    name: str
    default: typing.Any | EmptyType = Empty
    default_factory: typing.Any | EmptyType = Empty
    metadata: dict[str, Any] = dataclasses.field(default_factory=lambda: {})


@dataclasses.dataclass
class DataclassField(Field):
    @classmethod
    def collect(cls, typ: type) -> list[Self]:
        fields: list[Self] = []
        for f in dataclasses.fields(typ):  # pyright: ignore
            if not f.init:
                continue
            field = cls(
                name=f.name,
                default=f.default if f.default is not dataclasses.MISSING else Empty,
                default_factory=f.default_factory
                if f.default_factory is not dataclasses.MISSING
                else Empty,
                metadata=dict(f.metadata),
            )
            fields.append(field)
        return fields


@dataclasses.dataclass
class AttrsField(Field):
    @classmethod
    def collect(cls, typ: type) -> list[Self]:
        fields: list[Self] = []
        for f in typ.__attrs_attrs__:  # type: ignore
            if hasattr(f.default, "factory"):  # pyright: ignore
                default = None
                default_factory: Any = f.default.factory  # pyright: ignore
            else:
                default = f.default  # pyright: ignore
                default_factory = None
            field = cls(
                name=f.name,  # pyright: ignore
                default=default or Empty,  # pyright: ignore
                default_factory=default_factory or Empty,  # pyright: ignore
                metadata=f.metadata,  # pyright: ignore
            )
            fields.append(field)
        return fields


if TYPE_CHECKING:

    @dataclasses.dataclass
    class MsgspecField(Field):
        @classmethod
        def collect(cls, typ: type) -> list[Self]:
            return []

else:

    @dataclasses.dataclass
    class MsgspecField(Field):
        @classmethod
        def collect(cls, typ: type) -> list[Self]:
            import msgspec  # pyright: ignore

            fields: list[Self] = []
            for f in msgspec.structs.fields(typ):
                default = f.default if f.default is not msgspec.NODEFAULT else Empty
                default_factory = (
                    f.default_factory
                    if f.default_factory is not msgspec.NODEFAULT
                    else Empty
                )
                field = cls(
                    name=f.name,
                    default=default,
                    default_factory=default_factory,
                )
                fields.append(field)
            return fields


@dataclasses.dataclass
class PydanticV1Field(Field):
    @classmethod
    def collect(cls, typ: type) -> list[Self]:
        fields: list[Self] = []
        callable_view = CallableView.from_callable(typ, include_extras=True)
        for param in callable_view.parameters:
            name = param.name
            f = typ.__fields__[name]  # type: ignore

            field = cls(
                name=name,
                default=f.default  # pyright: ignore
                if f.default.__repr__() != "PydanticUndefined"  # pyright: ignore
                else Empty,
                default_factory=f.default_factory or Empty,  # pyright: ignore
            )
            fields.append(field)
        return fields


@dataclasses.dataclass
class PydanticV2Field(Field):
    @classmethod
    def collect(cls, typ: type) -> list[Self]:
        fields: list[Self] = []
        for name, f in typ.model_fields.items():  # type: ignore
            field = cls(
                name=name,  # pyright: ignore
                default=f.default  # pyright: ignore
                if f.default.__repr__() != "PydanticUndefined"  # pyright: ignore
                else Empty,
                default_factory=f.default_factory or Empty,  # pyright: ignore
            )
            fields.append(field)
        return fields


@dataclasses.dataclass
class PydanticV2DataclassField(Field):
    @classmethod
    def collect(cls, typ: type) -> list[Self]:
        fields: list[Self] = []
        for name, f in typ.__pydantic_fields__.items():  # type: ignore
            field = cls(
                name=name,  # pyright: ignore
                default=f.default or Empty,  # pyright: ignore
                default_factory=f.default_factory or Empty,  # pyright: ignore
            )
            fields.append(field)
        return fields


def fields(cls: type):
    class_type = ClassTypes.from_cls(cls)
    if class_type is None:
        raise ValueError(
            f"'{cls.__qualname__}' is not a currently supported kind of class. "
            "Must be one of: dataclass, pydantic, or attrs class."
        )

    return class_type.value.collect(cls)  # pyright: ignore


class ClassTypes(Enum):
    attrs = AttrsField
    dataclass = DataclassField
    pydantic_v1 = PydanticV1Field
    pydantic_v2 = PydanticV2Field
    pydantic_v2_dataclass = PydanticV2DataclassField
    msgspec = MsgspecField

    @classmethod
    def from_cls(cls, obj: type) -> ClassTypes | None:
        if hasattr(obj, "__pydantic_fields__"):
            return cls.pydantic_v2_dataclass

        if dataclasses.is_dataclass(obj):
            return cls.dataclass

        if hasattr(obj, "__struct_config__"):  # pyright: ignore
            assert obj.__struct_config__.__class__.__module__.startswith("msgspec")  # pyright: ignore
            return cls.msgspec

        try:
            import pydantic  # pyright: ignore
            from pydantic import BaseModel  # pyright: ignore
        except ImportError:  # pragma: no cover
            pass
        else:
            try:
                is_base_model = isinstance(obj, type) and issubclass(obj, BaseModel)
            except TypeError:  # pragma: no cover
                is_base_model = False

            if is_base_model:
                if getattr(pydantic, "__version__", "1.0").startswith("1."):  # pyright: ignore, pragma: no cover
                    return cls.pydantic_v1
                return cls.pydantic_v2

        if hasattr(obj, "__attrs_attrs__"):  # pyright: ignore
            return cls.attrs

        return None


def extract_dataclass_metadata(field: Field, cls: type[T]) -> list[T]:
    field_metadata = field.metadata.get("cappa")
    if not field_metadata:
        return []

    if not isinstance(field_metadata, cls):
        return []

    return [field_metadata]


def get_command_capable_object(obj: Any) -> type:
    """Convert raw functions into a stub class.

    Internally, a dataclass is constructed with a `__call__` method which **splats
    the arguments to the dataclass into the original callable.
    """
    if inspect.isfunction(obj):
        from cappa import Dep
        from cappa.command import Command

        function_args: list[Any] = []

        @functools.wraps(obj)
        def call(self: Any, *args: Any, **deps: Any):
            kwargs: dict[str, Any] = dataclasses.asdict(self)
            return obj(*args, **kwargs, **deps)

        callable_view = CallableView.from_callable(obj, include_extras=True)

        # We need to create a fake signature for the above callable, which does
        # not retain the `Arg` annotations
        signature = callable_view.signature
        sig_params: dict[str, inspect.Parameter] = dict(signature.parameters)
        signature._parameters = sig_params  # type: ignore
        call.__signature__ = signature  # type: ignore

        for param_view in callable_view.parameters:
            if not param_view.has_annotation:
                continue

            if find_annotations(
                param_view.type_view, Dep
            ) or param_view.type_view.is_subclass_of((Output, State, Command)):
                continue

            sig_params.pop(param_view.name, None)
            function_args.append(
                (
                    param_view.name,
                    param_view.type_view.raw,
                    dataclasses.field(
                        default=param_view.default
                        if param_view.has_default
                        else dataclasses.MISSING
                    ),
                )
            )

        result = dataclasses.make_dataclass(
            obj.__name__,
            function_args,
            namespace={"__call__": call},
        )
        result.__doc__ = obj.__doc__
        result.__cappa__ = getattr(obj, "__cappa__", None)  # type: ignore
        return result

    method_subcommands = collect_method_subcommands(obj)
    if method_subcommands:
        from cappa.subcommand import Subcommand

        kw_only: dict[str, typing.Any] = {}
        if sys.version_info >= (3, 10):
            kw_only["kw_only"] = True

        return dataclasses.make_dataclass(
            obj.__name__,
            [
                (
                    "__cappa_subcommand__",
                    Annotated[
                        typing.Union[method_subcommands],  # pyright: ignore
                        Subcommand("command", required=True),
                    ],
                    dataclasses.field(
                        repr=False, compare=False, default=None, **kw_only
                    ),
                ),
            ],
            bases=(obj,),
        )

    return obj


def collect_method_subcommands(cls: type) -> tuple[typing.Callable[..., Any], ...]:
    return tuple(
        method
        for _, method in inspect.getmembers(cls, callable)
        if hasattr(method, "__cappa__")
    )


def has_command(obj: object) -> bool:
    return hasattr(obj, "__cappa__")


def get_command(obj: type) -> Command[Any] | None:
    return getattr(obj, "__cappa__", None)



from __future__ import annotations

import contextlib
import enum
import functools
import inspect
import types
from datetime import date, datetime, time
from typing import (
    Any,
    BinaryIO,
    Callable,
    Final,
    Generator,
    List,
    Sequence,
    TextIO,
    Type,
    Union,
    cast,
)

from cappa.file_io import FileMode
from cappa.output import Exit
from cappa.state import S, State
from cappa.type_view import TypeView
from cappa.typing import T

__all__ = [
    "parse_list",
    "parse_literal",
    "parse_set",
    "parse_tuple",
    "parse_union",
    "parse_value",
    "unpack_arguments",
]


type_priority: Final = types.MappingProxyType(
    {
        None: 0,
        ...: 1,
        float: 2,
        int: 3,
        bool: 4,
        str: 5,
    }
)

Parser = Callable[..., T]
MaybeTypeView = Union[Type[T], TypeView[Type[T]]]


def unpack_arguments(value: object, type_view: TypeView[T]) -> Parser[T]:
    """`parse=` compatible function that splats values into a dataclass-like object constructor.

    For example, some `foo: Annotated[Object, Arg(parse[json.loads, splat_arguments])]` annotation
    would unpack the result of `json.loads` into the constructor of `Object`, like `Object(**data)`.
    """
    origin = type_view.strip_optional().fallback_origin
    mapper = origin

    value_type_view = TypeView(type(value))
    if value_type_view.is_mapping:
        return origin(**value)  # pyright: ignore

    if value_type_view.is_collection and not value_type_view.is_subclass_of(str):
        return origin(*value)  # pyright: ignore

    return mapper(value)


def _as_type_view(typ: type[T] | TypeView[type[T]]) -> TypeView[type[T]]:
    if isinstance(typ, TypeView):
        return typ
    return TypeView(typ)


def default_parse(value: str, type_view: TypeView[Any]) -> Any:
    """Perform the default inferred parse behavior of cappa's type inference system."""
    parser: Parser[Any] = parse_value(type_view)
    return parser(value)


def parse_value(typ: MaybeTypeView[T]) -> Parser[T]:
    """Create a value parser for the given annotation.

    Examples:
        >>> from typing import List, Literal, Tuple
        >>> list_parser = parse_value(List[int])
        >>> tuple_parser = parse_value(Tuple[int, ...])
        >>> literal_parser = parse_literal(Literal["foo"])
    """
    type_view = _as_type_view(typ)

    if type_view.is_type_alias:
        return parse_value(type_view.strip_type_alias())

    if type_view.is_literal:
        return parse_literal(type_view)

    if type_view.is_union:
        return parse_union(type_view)

    if type_view.is_subclass_of((str, bool, int, float)):
        return type_view.annotation

    if type_view.is_none_type:
        return parse_none  # type: ignore

    if type_view.is_subclass_of(enum.Enum):
        return parse_enum(type_view)

    if type_view.is_subclass_of(datetime):
        return datetime.fromisoformat  # type: ignore

    if type_view.is_subclass_of(date):
        return date.fromisoformat  # type: ignore

    if type_view.is_subclass_of(time):
        return time.fromisoformat  # type: ignore

    if type_view.is_subclass_of(list):
        return parse_list(type_view)  # type: ignore

    if type_view.is_subclass_of(set):
        return parse_set(type_view)  # type: ignore

    if type_view.is_subclass_of(tuple):
        return parse_tuple(type_view)  # type: ignore

    if type_view.is_subclass_of((TextIO, BinaryIO)):
        return parse_file_io(type_view)

    return parse_fallback(type_view.annotation)


def parse_fallback(fallback: type[T]) -> Parser[T]:
    def fallback_mapper(v: Any) -> T:
        if isinstance(v, fallback):
            return v

        return fallback(v)  # type: ignore

    return fallback_mapper


def parse_literal(typ: MaybeTypeView[T]) -> Parser[T]:
    """Create a value parser for a given literal value."""
    type_view = _as_type_view(typ)
    unique_type_args = set(type_view.args)

    def literal_mapper(value: T) -> T:
        if isinstance(value, (list, set, tuple)):
            for item in cast(Sequence[Any], value):
                if item not in unique_type_args:
                    raise choices_error(type_view.args, item)

            return cast(T, value)

        if value in unique_type_args:
            return value

        for type_arg in unique_type_args:
            raw_value = str(type_arg)
            if raw_value == value:
                return type_arg

        raise choices_error(type_view.args, value)

    return literal_mapper


def parse_enum(typ: MaybeTypeView[T]):
    type_view = _as_type_view(typ)
    choices: tuple[Any, ...] = tuple(v.value for v in type_view.annotation)

    def enum_mapper(value: Any) -> Any:
        try:
            return type_view.annotation(value)
        except ValueError:
            raise choices_error(choices, value)

    return enum_mapper


def parse_list(typ: MaybeTypeView[list[T]]) -> Parser[list[T]]:
    """Create a value parser for a list of given type `of_type`."""
    type_view = _as_type_view(typ)
    inner_mapper: Parser[T] = parse_value(type_view.inner_types[0])

    def list_mapper(value: list[Any]) -> list[T]:
        return [inner_mapper(v) for v in value]

    return list_mapper


def parse_set(typ: MaybeTypeView[list[T]]) -> Parser[set[T]]:
    """Create a value parser for a list of given type `of_type`."""
    type_view = _as_type_view(typ)
    inner_mapper: Parser[T] = parse_value(type_view.inner_types[0])

    def set_mapper(value: list[Any]) -> set[T]:
        return {inner_mapper(v) for v in value}

    return set_mapper


def parse_tuple(typ: MaybeTypeView[T]) -> Parser[tuple[T]]:
    """Create a value parser for a tuple with type-args of given `type_args`."""
    type_view = _as_type_view(typ)
    if type_view.is_variadic_tuple:
        assert type_view.args
        inner_type = type_view.args[0]
        list_mapper = parse_list(List[inner_type])  # type: ignore

        def unbounded_tuple_mapper(value: list[Any]) -> tuple[Any, ...]:
            return tuple(list_mapper(value))

        return unbounded_tuple_mapper

    def tuple_mapper(value: list[Any]) -> tuple[Any]:
        result: list[Any] = []
        for inner_type, inner_value in zip(type_view.inner_types, value):
            inner_mapper: Parser[T] = parse_value(inner_type)
            inner_value = inner_mapper(inner_value)
            result.append(inner_value)
        return tuple(result)

    return tuple_mapper


def parse_union(typ: MaybeTypeView[T]) -> Parser[T]:
    """Create a value parser for a Union with type-args of given `type_args`."""

    def type_priority_key(type_type_view: TypeView[Any]) -> int:
        return type_priority.get(type_type_view.annotation, 1)

    def union_mapper(value: Any) -> Any:
        exceptions: list[str] = []
        for mapper_type_view, mapper in mappers:
            try:
                return mapper(value)
            except (ValueError, TypeError) as e:
                if mapper is parse_none:
                    err = " - <no value>"
                else:
                    err = f" - {mapper_type_view.repr_type}: {e}"
                exceptions.append(err)

        # Perhaps we should be showing all failed mappings at some point. As-is,
        # the preferred interpretation will be determined by order in the event
        # of all mappings failing
        reasons = "\n".join(exceptions)
        raise ValueError(f"Possible variants\n{reasons}")

    type_view = _as_type_view(typ)
    mappers: list[tuple[TypeView[T], Callable[..., Any]]] = [
        (t, parse_value(t))
        for t in sorted(type_view.inner_types, key=type_priority_key)
    ]

    return union_mapper


def parse_none(value: Any) -> None:
    """Create a value parser for None."""
    if value is None:
        return
    raise ValueError(value)


def parse_file_io(typ: MaybeTypeView[T]) -> Parser[T]:
    type_view = _as_type_view(typ)

    def file_io_mapper(value: str) -> T:
        try:
            file_mode: FileMode = next(
                f for f in type_view.metadata if isinstance(f, FileMode)
            )
        except StopIteration:
            file_mode = FileMode()

            if type_view.is_subclass_of(BinaryIO):
                file_mode.mode += "b"

        return file_mode(value)  # type: ignore

    return file_io_mapper


def evaluate_parse(
    parsers: Parser[T] | Sequence[Parser[Any]],
    type_view: TypeView[T],
    state: State[S] | None = None,
) -> Callable[..., Any]:
    import inspect

    from cappa.invoke.base import fulfill_deps

    state = State.ensure(state)  # type: ignore

    if callable(parsers):
        parsers = [parsers]

    filled_parsers = [
        functools.partial(
            parser,
            **fulfill_deps(
                parser,
                {TypeView: type_view, State: state},
                allow_empty=True,
            ).kwargs,
        )
        for parser in parsers
    ]

    if len(parsers) == 1:
        return filled_parsers[0]

    # Check if any parser is async
    has_async_parser = any(inspect.iscoroutinefunction(p.func) for p in filled_parsers)

    if has_async_parser:

        async def async_sequence_parsers(value: Any) -> T:
            result = value
            for parser in filled_parsers:
                result = parser(result)
                # If the result is a coroutine, await it
                if inspect.iscoroutine(result):
                    result = await result

            return result  # pyright: ignore[reportReturnType]

        return async_sequence_parsers

    def sequence_parsers(value: Any) -> T:
        result = value
        for parser in filled_parsers:
            result = parser(result)

        return result

    return sequence_parsers


def choices_error(choices: Sequence[Any], value: Any) -> Exception:
    options = ", ".join(f"{t!r}" for t in choices)
    return ValueError(f"Invalid choice: '{value}' (choose from {options})")


@contextlib.contextmanager
def apply_parse(
    parse_fn: Callable[..., Any],
    prog: str,
    value: Any,
    is_parsed: bool,
    names_str: str,
) -> Generator[Any, None, None]:
    if not is_parsed:
        try:
            value = parse_fn(value)
        except Exception as e:
            raise Exit(
                f"Invalid value for '{names_str}': {e}",
                code=2,
                prog=prog,
            )
    yield value


def parse_handler(
    parse_fn: Callable[..., Any],
    prog: str,
    value: Any,
    names_str: str,
) -> Callable[[Any, bool], Any]:
    is_async_value = inspect.iscoroutine(value)
    is_async_parse = inspect.iscoroutinefunction(
        parse_fn
    ) or inspect.isasyncgenfunction(parse_fn)

    if is_async_value or is_async_parse:

        async def async_parse(raw_value: Any, is_parsed: bool) -> Any:
            if inspect.iscoroutine(raw_value):
                raw_value = await raw_value

            with apply_parse(parse_fn, prog, raw_value, is_parsed, names_str) as parsed:
                if inspect.iscoroutine(parsed):
                    try:
                        return await parsed
                    except Exception as e:
                        raise Exit(
                            f"Invalid value for '{names_str}': {e}",
                            code=2,
                            prog=prog,
                        )
                return parsed

        return async_parse

    def sync_parse(raw_value: Any, is_parsed: bool) -> Any:
        with apply_parse(parse_fn, prog, raw_value, is_parsed, names_str) as parsed:
            return parsed

    return sync_parse



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Union

import pytest

from tests.utils import Backend, backends, parse


@backends
def test_valid_tagged_unions(backend: Backend):
    @dataclass
    class ArgTest:
        name: Union[
            tuple[Literal["one"], str],
            tuple[Literal["two"], int],
            tuple[
                Literal["three"],
                float,
            ],
        ]

    test = parse(ArgTest, "one", "string", backend=backend)
    assert test.name == ("one", "string")

    test = parse(ArgTest, "two", "4", backend=backend)
    assert test.name == ("two", 4)

    test = parse(ArgTest, "three", "1.4", backend=backend)
    assert test.name == ("three", 1.4)


@backends
def test_disallowed_different_arity_variants(backend: Backend):
    @dataclass
    class ArgTest:
        name: Union[tuple[str, str], tuple[str, str, str]]

    with pytest.raises(ValueError) as e:
        parse(ArgTest, "one", "string", backend=backend)

    assert (
        str(e.value).lower().replace("typing.", "")
        == "on field 'name', mismatch of arity between union variants. `tuple[str, str]` produces `num_args=2`, `tuple[str, str, str]` produces `num_args=3`."
    )



from __future__ import annotations

import contextlib
import os
import sys
from dataclasses import dataclass
from typing import Any, Union
from unittest.mock import patch

import pytest
from typing_extensions import Unpack

from cappa import argparse, parser
from cappa.base import Backend
from cappa.command import Command
from cappa.invoke.types import DepTypes, InvokeCallable
from cappa.output import Exit
from cappa.testing import CommandRunner, RunnerArgs, T

__all__ = [
    "Backend",
    "CapsysOutput",
    "backends",
    "invoke",
    "invoke_async",
    "parse",
    "require_dataclass_kw_only",
    "runner",
]

require_dataclass_kw_only = pytest.mark.skipif(
    sys.version_info < (3, 10), reason="kw_only requires Python 3.10+"
)

backends = pytest.mark.parametrize(
    "backend",
    [
        None,
        pytest.param(argparse.backend, marks=pytest.mark.argparse),
    ],
)

runner: CommandRunner[Any] = CommandRunner(base_args=[])


def parse(
    cls: type[T] | InvokeCallable[T] | Command[T],
    *args: str,
    **kwargs: Unpack[RunnerArgs],
) -> T:
    return runner.parse(cls, *args, **kwargs)


def invoke(
    cls: type[T] | InvokeCallable[T] | Command[T],
    *args: str,
    deps: DepTypes | None = None,
    **kwargs: Unpack[RunnerArgs],
):
    return runner.invoke(cls, *args, deps=deps, **kwargs)


def invoke_async(
    cls: type[T] | InvokeCallable[T] | Command[T],
    *args: str,
    deps: DepTypes | None = None,
    **kwargs: Unpack[RunnerArgs],
):
    return runner.invoke_async(cls, *args, deps=deps, **kwargs)


def parse_completion(
    cls: type, *args: str, location: str | None = None
) -> Union[str, None]:
    env = {
        "COMPLETION_LINE": " ".join(["test.py", *args]),
        "COMPLETION_LOCATION": location if location is not None else len(args) + 1,
    }
    with patch("os.environ", new=env):
        with pytest.raises(Exit) as e:
            parse(cls, "--completion", "complete", backend=parser.backend)

        assert e.value.code == 0
        if e.value.message:
            return str(e.value.message)
        return None


@contextlib.contextmanager
def ignore_docstring_parser(monkeypatch: Any):
    import importlib

    cappa_command = importlib.import_module("cappa.docstring")

    with monkeypatch.context() as m:
        m.setattr(cappa_command, "docstring_parser", None)
        yield


@contextlib.contextmanager
def terminal_width(width: int = 1000):
    env = os.environ.copy()
    env["COLUMNS"] = str(width)
    with patch("os.environ", new=env), patch("rich.console.Console._environ", new=env):
        yield


def strip_trailing_whitespace(text: str):
    return "\n".join([line.rstrip() for line in text.split("\n")])


@dataclass
class CapsysOutput:
    stdout: str
    stderr: str

    @classmethod
    def from_capsys(cls, capsys: Any):
        outerr = capsys.readouterr()

        out = strip_trailing_whitespace(outerr.out)
        err = strip_trailing_whitespace(outerr.err)
        return cls(out, err)



from __future__ import annotations

import asyncio
import contextlib
from dataclasses import dataclass
from typing import Any

import cappa
from tests.utils import Backend, backends


@dataclass
class Command:
    value: str = "test"
    closed: bool = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args: Any):
        object.__setattr__(self, "closed", True)


@backends
def test(backend: Backend):
    async def run_test():
        async with contextlib.AsyncExitStack() as stack:
            result = await cappa.parse_async(
                Command, argv=[], backend=backend, exit_stack=stack
            )

            assert not result.closed
            assert result.value == "test"

            return result

    result = asyncio.run(run_test())
    assert result.closed



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


async def async_fetch_raw_value() -> str:
    await asyncio.sleep(0.001)
    return "10"


async def async_parse_multiply(value: str) -> int:
    await asyncio.sleep(0.001)
    return int(value) * 2


@dataclass
class Command:
    value: Annotated[
        int,
        cappa.Arg(
            default=cappa.ValueFrom(async_fetch_raw_value),
            parse=async_parse_multiply,
        ),
    ]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    # ValueFrom returns is_parsed=True, so parse is NOT applied
    result = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert result.value == "10"

    # With explicit value: parse IS applied
    result = asyncio.run(cappa.invoke_async(Command, argv=["5"], backend=backend))
    assert result.value == 10



#!/usr/bin/env python

from __future__ import annotations

import enum
import logging
import sqlite3
from dataclasses import dataclass, field
from typing import Literal, Union

from typing_extensions import Annotated

import cappa

log = logging.getLogger(__name__)


class OtherOptions(enum.Enum):
    a = "a"
    b = "b"
    c = "c"


@dataclass
class Example:
    """Example program.

    Args:
       name: Required name argument
       flag: optional flag to enable
    """

    name: str
    value: Annotated[int, cappa.Arg(short="-V", long="--val", default=0)]

    options: Annotated[
        Union[Literal["one"], Literal["two"], Literal[3]], cappa.Arg(short=True)
    ]
    moptions: Annotated[OtherOptions, cappa.Arg(short=True)]

    subcommand: Annotated[Union[MeowCommand, BarkCommand], cappa.Subcommand]

    flags: Annotated[list[str], cappa.Arg(short=True, long=True)] = field(
        default_factory=lambda: []
    )
    flag: bool = False  # --flag


def db() -> sqlite3.Connection:
    return sqlite3.connect("sqlite:///")


def meow(
    command: Example,
    meow: MeowCommand,
    db: Annotated[sqlite3.Connection, cappa.Dep(db)],
):
    result = db.execute(f"select '{command.name}', {meow.times} + 1").fetchone()
    log.info(result)


@cappa.command(name="meow", invoke=meow)
@dataclass
class MeowCommand:
    times: int


def bark(bark: BarkCommand):
    log.info(bark)


@cappa.command(name="bark", invoke=bark)
@dataclass
class BarkCommand: ...


# invoke cli parsing
def main(argv: list[str] | None = None):
    logging.basicConfig()

    args: Example = cappa.parse(Example, argv=argv, version="1.2.3")
    log.info(args)

    cappa.invoke(Example, argv=argv, version="1.2.3")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_positional_arg(backend: Backend):
    @dataclass
    class ArgTest:
        numbers: tuple[int, ...]

    test = parse(ArgTest, "1", "2", "3", "4", backend=backend)
    assert test.numbers == (1, 2, 3, 4)


@backends
def test_option_flag(backend: Backend):
    @dataclass
    class ArgTest:
        numbers: Annotated[tuple[int, ...], cappa.Arg(short=True)]

    test = parse(ArgTest, "-n", "1", "-n", "2", "-n", "3", "-n", "4", backend=backend)
    assert test.numbers == (1, 2, 3, 4)



from __future__ import annotations

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Hashable,
    Protocol,
    Type,
    TypeVar,
    Union,
    overload,
)

from cappa.state import S, State

if TYPE_CHECKING:
    from cappa.command import Command, FinalCommand
    from cappa.invoke.types import InvokeCallable, Resolved
    from cappa.output import Output


T = TypeVar("T")
U = TypeVar("U")

CappaCapable = Union["InvokeCallable[T]", Type[T], "Command[T]"]


class Backend(Protocol):
    def __call__(
        self,
        command: FinalCommand[T],
        argv: list[str],
        output: Output,
        prog: str,
        provide_completions: bool = False,
    ) -> tuple[Any, Command[T], dict[str, Any]]: ...  # pragma: no cover


class FuncOrClassDecorator(Protocol):
    @overload
    def __call__(self, x: type[T], /) -> type[T]: ...
    @overload
    def __call__(self, x: T, /) -> T: ...


@dataclass()
class ParseResult(Generic[T, S]):
    """Result of parsing a command with all collected metadata.

    Attributes:
        root_command: The root command that was parsed.
        parsed_command: The selected command (may be a subcommand if one was invoked).
        instance: The deferred instantiated command object.
        implicit_deps: Mapping of command classes to their deferred instances, collected during parsing.
        output: The output handler for the command.
        state: The state object for the command.
    """

    root_command: Command[T]
    parsed_command: Command[T]
    instance: Resolved[T]
    implicit_deps: dict[Hashable, Any]
    state: State[S]
    output: Output



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, CapsysOutput, backends, invoke


@backends
def test_outputs_output(backend: Backend, capsys: Any):
    @dataclass
    class TopLevelCommand:
        def __call__(self, out: cappa.Output):
            out.output("woah!")
            out.error("woops!")

    invoke(TopLevelCommand, backend=backend)

    outerr = capsys.readouterr()
    assert outerr.out == "woah!\n"
    assert outerr.err == "Error: woops!\n"


@backends
def test_output_callable(backend: Backend, capsys: Any):
    @dataclass
    class TopLevelCommand:
        def __call__(self, out: cappa.Output):
            out("woah!")

    invoke(TopLevelCommand, backend=backend)

    out = CapsysOutput.from_capsys(capsys)
    out.stdout = "woah!\n"
    out.stderr = ""


def _raise_value_error(value: str) -> str:
    raise ValueError("nope")


@backends
def test_invoke_routes_parse_errors_through_output(backend: Backend, capsys: Any):
    """A parse= callback that raises should produce a visible error on stderr.

    Regression test: previously `invoke()` (and `invoke_async`/`parse`) did not
    pass `output=` into the resolution of `parse_result.instance`, so any
    `cappa.Exit` raised from an `Arg(parse=...)` callback would be caught by
    `Resolved.handle_exit` with `output=None` and silently re-raised, producing
    a non-zero exit with no message.
    """

    @dataclass
    class Cmd:
        x: Annotated[
            str,
            cappa.Arg(parse=_raise_value_error, parse_inference=False, long="--x"),
        ] = "default"

        def __call__(self) -> None:
            pass

    with pytest.raises(cappa.Exit) as e:
        invoke(Cmd, "--x", "foo", backend=backend)

    assert e.value.code == 2
    out = CapsysOutput.from_capsys(capsys)
    assert "Invalid value for '--x': nope" in out.stderr



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import patch

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_missing_fallback(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[str, cappa.Arg(default=cappa.Env("ASDF", default="wat"))]

    test = parse(ArgTest, backend=backend)
    assert test.default == "wat"


@backends
def test_has_env_var(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[str, cappa.Arg(default=cappa.Env("ASDF", default="wat"))]

    with patch("os.environ", new={"ASDF": "asdf!"}):
        test = parse(ArgTest, backend=backend)
    assert test.default == "asdf!"


@backends
def test_env_defers_to_real_value(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[str, cappa.Arg(default=cappa.Env("ASDF", default="wat"))]

    test = parse(ArgTest, "test", backend=backend)
    assert test.default == "test"


@backends
def test_mapping_is_applied(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[bool, cappa.Arg(default=cappa.Env("ASDF", default="wat"))]

    test = parse(ArgTest, backend=backend)
    assert test.default is True


@backends
def test_class_default_gets_applied_to_env_default(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[str, cappa.Arg(default=cappa.Env("ASDF"))] = "wat"

    test = parse(ArgTest, backend=backend)
    assert test.default == "wat"


@backends
def test_class_default_still_abides_env_var(backend: Backend):
    @dataclass
    class ArgTest:
        default: Annotated[str, cappa.Arg(default=cappa.Env("ASDF"))] = "wat"

    with patch("os.environ", new={"ASDF": "foo"}):
        test = parse(ArgTest, backend=backend)
    assert test.default == "foo"



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Literal, Union

import pytest
from typing_extensions import Annotated, Doc

import cappa
from cappa.output import Exit
from tests.utils import Backend, backends, parse


@pytest.mark.help
@backends
def test_explicit_parse_function(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        numbers: Annotated[int, cappa.Arg(parse=int, help="example")]

    with pytest.raises(Exit):
        parse(ArgTest, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert re.match(r".*NUMBERS\s+example.*", stdout, re.DOTALL)


@pytest.mark.help
@backends
def test_choices_in_help(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        numbers: Annotated[
            Union[Literal[1], Literal[2]], cappa.Arg(parse=int, help="example")
        ]

    result = parse(ArgTest, "1", backend=backend)
    assert result == ArgTest(1)

    with pytest.raises(Exit):
        parse(ArgTest, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert "Valid options: 1, 2" in stdout


@pytest.mark.help
@backends
def test_pep_727_doc_annotated(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        """Test.

        Arguments:
            doc_beats_docstring: loses.
        """

        only_doc: Annotated[int, cappa.Arg(parse=int), Doc("Use Doc if exists")]
        doc_beats_docstring: Annotated[
            int, cappa.Arg(parse=int), Doc("Doc beats docstring")
        ]
        prefer_arg: Annotated[
            int, cappa.Arg(parse=int, help="Arg wins"), Doc("Doc loses")
        ]

    with pytest.raises(Exit):
        parse(ArgTest, "--help", backend=backend)

    stdout = capsys.readouterr().out
    assert "Use Doc if exists" in stdout
    assert "Doc beats docstring" in stdout

    assert "Arg wins" in stdout
    assert "Doc loses" not in stdout



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

import pytest
from typing_extensions import Annotated

import cappa
from cappa.parser import Value
from tests.utils import Backend, backends, parse


################################
def exit():
    raise cappa.Exit("message")


@backends
def test_callable_action_fast_exits(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(action=exit, short=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "-v", "the", "rest", "is", "trash", backend=backend)
    assert e.value.message == "message"


################################
def truncate(value: Value[str]):
    return value.value[1:]


@backends
def test_uses_return_value(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(action=truncate, short=True)]

    args = parse(Args, "-v", "okay", backend=backend)
    assert args.value == "kay"


@backends
def test_custom_arg(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[str, cappa.Arg(action=truncate)]

    args = parse(Args, "okay", backend=backend)
    assert args.value == "kay"


################################
def command_name(command: cappa.Command[Any]):
    return command.real_name()


@dataclass
class SubBSub:
    value: Annotated[str, cappa.Arg(action=command_name, short=True)]


@dataclass
class SubA:
    value: Annotated[str, cappa.Arg(action=command_name, short=True)]


@dataclass
class SubB:
    cmd: cappa.Subcommands[SubBSub]


@backends
def test_subcommand_name(backend: Backend):
    @dataclass
    class Args:
        cmd: cappa.Subcommands[Union[SubA, SubB]]

    args = parse(Args, "sub-a", "-v", "one", backend=backend)
    assert isinstance(args.cmd, SubA)
    assert args.cmd.value == "sub-a"

    args = parse(Args, "sub-b", "sub-b-sub", "-v", "one", backend=backend)
    assert isinstance(args.cmd, SubB)
    assert args.cmd.cmd.value == "sub-b-sub"


################################
def custom_out(out: cappa.Output):
    out.output("woah")
    return 1


@backends
def test_custom_action_output_dep(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        value: Annotated[int, cappa.Arg(action=custom_out, short=True)]

    args = parse(Args, "-v", "one", backend=backend)
    assert args.value == 1

    out = capsys.readouterr().out
    assert out == "woah\n"



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse_completion


def test_long_option_name():
    @dataclass
    class Args:
        default: bool = False

    result = parse_completion(Args, "--d")
    assert result == "--default:<default> (Default: False)"


def test_long_option_name_with_help():
    @dataclass
    class Args:
        default: Annotated[bool, cappa.Arg(help="Enables default")] = False

    result = parse_completion(Args, "--d")
    assert result == "--default:Enables default (Default: False)"


def test_multiple_matches():
    @dataclass
    class Args:
        apple: Annotated[str, cappa.Arg(short=True, help="apple")]
        banana: Annotated[str, cappa.Arg(short=True, help="banana")]

    result = parse_completion(Args, "-")
    assert result
    assert "-a:apple\n-b:banana" in result


def test_short_name():
    @dataclass
    class Args:
        apple: Annotated[str, cappa.Arg(short=True)]

    result = parse_completion(Args, "-a")
    assert result is None


def test_no_help_value_name():
    @dataclass
    class Args:
        apple: Annotated[str, cappa.Arg(long=True)]

    result = parse_completion(Args, "--a")
    assert result == "--apple:<apple>"



from cappa.completion.types import ShellHandler

shells = [
    ShellHandler(
        "bash",
        template="""\
_%(safe_prog_name)s_completion() {
    local IFS=$'\\n' local response
    response=$(env COMPLETION_LINE="${COMP_WORDS[*]}" COMPLETION_LOCATION=$((COMP_CWORD+1)) $1 --completion=complete)

    for completion in $response; do
        IFS=':' read value help <<< "$completion"

        if [[ $type == 'file' ]]; then
            COMPREPLY=()
            compopt -o default
        else
            COMPREPLY+=($value)
        fi
    done
    return 0
}

_%(safe_prog_name)s_completion_setup() {
    complete -o nosort -F _%(safe_prog_name)s_completion %(prog_name)s
}

_%(safe_prog_name)s_completion_setup;
        """,
    ),
    ShellHandler(
        "zsh",
        template="""\
#compdef %(prog_name)s

_%(safe_prog_name)s_completion() {
    local -a completions
    (( ! $+commands[%(prog_name)s] )) && return 1

    completions=("${(@f)$(env COMPLETION_LINE="${words[*]}" COMPLETION_LOCATION=$((CURRENT)) \
%(prog_name)s %(completion_arg)s=complete)}")

    if [[ -n "$completions" ]]; then
        if [[ "${completions[1]}" == "file" ]]; then
            _files
        else
            _describe 'options' completions
        fi
    fi
}

if [[ $zsh_eval_context[-1] == loadautofunc ]]; then
    _%(safe_prog_name)s_completion "$@"
else
    compdef _%(safe_prog_name)s_completion %(prog_name)s
fi
""",
    ),
    ShellHandler(
        "fish",
        template="""\
function _%(safe_prog_name)s_completion
    set -l response (env COMPLETION_LINE=(commandline -cp) COMPLETION_LOCATION=(math (commandline -C) + 1) \
%(prog_name)s %(completion_arg)s=complete)

    for completion in $response
        set -l value (string split --max 1 ":" -- $completion)

        if test $value[1] = "file"
            __fish_complete_path
        else
            echo $value[2]
        end
    end
end

complete --no-files --command %(prog_name)s --arguments "(_%(safe_prog_name)s_completion)"
""",
    ),
]
available_shells = {s.name: s for s in shells}



from dataclasses import dataclass
from typing import Sequence, Union
from unittest.mock import Mock, patch

import pytest
from docutils import nodes
from docutils.core import publish_from_doctree  # pyright: ignore
from docutils.parsers.rst.states import Body, RSTStateMachine
from docutils.utils import new_document
from typing_extensions import Annotated

import cappa
from cappa.ext.docutils import CappaDirective


@dataclass
class Foo:
    bar: int
    baz: Annotated[int, cappa.Arg(help="asdf")]


@cappa.command(help="has help", description="has description")
@dataclass
class Bar:
    bar: int = 4


@dataclass
class Subcommand:
    subcmd: cappa.Subcommands[Union[Bar, Foo]]


def create_directive(*, style: str, cls_name: str = "Foo", terminal_width: int = 0):
    with patch("docutils.parsers.rst.states.Body.build_table"):
        state = Body(RSTStateMachine([], ""))

    state.document = new_document("<rst-doc>", None)
    return CappaDirective(
        "cappa",
        [f"tests.ext.test_docutils.{cls_name}"],
        {"style": style, "terminal-width": terminal_width},
        [],  # type: ignore
        0,
        0,
        "",
        state,
        Mock(),
    )


def render(nodes: Sequence[nodes.Node]) -> str:
    doc = new_document("<rst-doc>", None)
    doc += nodes
    return publish_from_doctree(doc, writer_name="html").decode()  # pyright: ignore


def test_invalid():
    directive = create_directive(style="wat")
    with pytest.raises(ValueError):
        directive.run()


def test_terminal():
    directive = create_directive(style="terminal")
    nodes = directive.run()
    result = render(nodes)
    assert "Usage: foo" in result
    assert "BAR" in result
    assert "Show this message and exit." in result
    assert "asdf" in result


def test_native():
    directive = create_directive(style="native")
    nodes = directive.run()
    result = render(nodes)
    assert "<h1>foo</h1>" in result
    assert '<span class="section-subtitle">Arguments</span>' in result
    assert '<tt class="first docutils literal">bar BAR</tt>' in result
    assert '<span class="section-subtitle">Help</span>' in result
    assert (
        '<span class="pre">--help</span></tt><span>: Show this message and exit.</span>'
        in result
    )
    assert "asdf" in result


def test_has_help_desc():
    directive = create_directive(style="native", cls_name="Bar")
    nodes = directive.run()
    result = render(nodes)
    assert "has help" in result
    assert "has description" in result


def test_subcommand():
    directive = create_directive(style="native", cls_name="Subcommand")
    nodes = directive.run()
    result = render(nodes)
    assert "subcommand" in result


def test_terminal_width():
    directive = create_directive(style="terminal", terminal_width=60)
    nodes = directive.run()
    render(nodes)



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class ArgTest:
    numbers: tuple[int, str, float]


@backends
def test_valid(backend: Backend):
    test = parse(ArgTest, "1", "2", "3.4", backend=backend)
    assert test.numbers == (1, "2", 3.4)


@backends
def test_tuple_option(backend: Backend):
    @dataclass
    class Example:
        start_project: Annotated[
            tuple[int, float],
            cappa.Arg(short=True, default=(1, 9), required=False),
        ]

    test = parse(Example, backend=backend)
    assert test == Example(start_project=(1, 9))

    test = parse(Example, "-s", "2", "2.4", backend=backend)
    assert test == Example(start_project=(2, 2.4))

    # Missing values
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "-s", "1", backend=backend)

    assert e.value.code == 2

    if backend:
        assert str(e.value.message).lower() == "argument -s: expected 2 arguments"
    else:
        assert (
            e.value.message == "Argument '-s' requires 2 values, found 1 ('1' so far)"
        )

    # Extra values
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "-s", "1", "2", "3", backend=backend)
    assert e.value.code == 2
    assert str(e.value.message).lower() == "unrecognized arguments: 3"


@backends
def test_optional_tuple_option(backend: Backend):
    @dataclass
    class Example:
        start_project: Annotated[
            tuple[int, float] | None,
            cappa.Arg(short=True, required=False),
        ] = None

    test = parse(Example, backend=backend)
    assert test == Example(start_project=None)

    test = parse(Example, "-s", "2", "2.4", backend=backend)
    assert test == Example(start_project=(2, 2.4))

    # Missing values
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "-s", "1", backend=backend)

    assert e.value.code == 2

    if backend:
        assert str(e.value.message).lower() == "argument -s: expected 2 arguments"
    else:
        assert (
            e.value.message == "Argument '-s' requires 2 values, found 1 ('1' so far)"
        )

    # Extra values
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "-s", "1", "2", "3", backend=backend)
    assert e.value.code == 2
    assert str(e.value.message).lower() == "unrecognized arguments: 3"



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Annotated, Literal

import cappa
from tests.utils import Backend, backends, parse


if sys.version_info >= (3, 12):
    type Numbers = Literal[1, 2, 3]  # pyright: ignore

    @backends
    def test_scalar(backend: Backend):
        @dataclass
        class ArgTest:
            value: Numbers

        test = parse(ArgTest, "1", backend=backend)
        assert test.value == 1

    @backends
    def test_list_numbers(backend: Backend):
        @dataclass
        class ArgTest:
            value: list[Numbers]

        test = parse(ArgTest, "1", backend=backend)
        assert test.value == [1]

    @backends
    def test_annotated(backend: Backend):
        @dataclass
        class ArgTest:
            value: Annotated[Numbers, cappa.Arg()]

        test = parse(ArgTest, "1", backend=backend)
        assert test.value == 1



from __future__ import annotations

import sys
import textwrap
from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import parse


def test_native_backend(capsys: Any):
    @dataclass
    class ArgTest:
        arg1: Annotated[str, cappa.Arg(deprecated=True)] = "default"
        arg2: Annotated[str, cappa.Arg(short="a", long="--aaaa", deprecated=True)] = (
            "default"
        )
        arg3: Annotated[str, cappa.Arg(short="b", deprecated="Use something else")] = (
            "default"
        )

    result = parse(ArgTest)
    assert result.arg1 == "default"
    assert result.arg2 == "default"
    assert result.arg3 == "default"
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "a")
    assert result.arg1 == "a"
    err = capsys.readouterr().err
    assert err == "Error: Argument `arg1` is deprecated\n"

    result = parse(ArgTest, "-a", "1")
    assert result.arg2 == "1"
    err = capsys.readouterr().err
    assert err == "Error: Option `-a` is deprecated\n"

    result = parse(ArgTest, "-b", "1")
    assert result.arg3 == "1"
    err = capsys.readouterr().err
    assert err == "Error: Option `-b` is deprecated: Use something else\n"


@pytest.mark.skipif(sys.version_info >= (3, 13), reason="requires python3.13 or higher")
def test_argparse_le_313(capsys: Any):
    """Below 3.13, this option has no effect."""

    @dataclass
    class ArgTest:
        arg1: Annotated[str, cappa.Arg(deprecated=True)] = "default"
        arg2: Annotated[str, cappa.Arg(short="a", long="--aaaa", deprecated=True)] = (
            "default"
        )
        arg3: Annotated[str, cappa.Arg(short="b", deprecated="Use something else")] = (
            "default"
        )

    result = parse(ArgTest, backend=cappa.argparse.backend)
    assert result.arg1 == "default"
    assert result.arg2 == "default"
    assert result.arg3 == "default"
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "1", "-a", "1", "-b", "1", backend=cappa.argparse.backend)
    assert result.arg1 == "1"
    assert result.arg1 == "1"
    assert result.arg1 == "1"
    err = capsys.readouterr().err
    assert err == ""


@pytest.mark.skipif(
    sys.version_info < (3, 13), reason="Below 3.13, the behavior is different"
)
def test_argparse_ge_313(capsys: Any):
    @dataclass
    class ArgTest:
        arg1: Annotated[str, cappa.Arg(deprecated=True)] = "default"
        arg2: Annotated[str, cappa.Arg(short="a", long="--aaaa", deprecated=True)] = (
            "default"
        )
        arg3: Annotated[str, cappa.Arg(short="b", deprecated="Use something else")] = (
            "default"
        )

    result = parse(ArgTest, backend=cappa.argparse.backend)
    assert result.arg1 == "default"
    assert result.arg2 == "default"
    assert result.arg3 == "default"
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "1", "-a", "1", "-b", "1", backend=cappa.argparse.backend)
    assert result.arg1 == "1"
    assert result.arg1 == "1"
    assert result.arg1 == "1"
    err = capsys.readouterr().err.replace("arg-test: ", "")
    assert err == textwrap.dedent(
        """\
        warning: argument 'arg1' is deprecated
        warning: option '-a' is deprecated
        warning: option '-b' is deprecated
        """
    )



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa import argparse
from tests.utils import Backend, backends, parse


@backends
def test_invalid_choice_help(backend: Backend):
    @dataclass
    class Args:
        arg: tuple[str, str]
        option: Annotated[str, cappa.Arg(long=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "arg", "--option=opt", backend=backend)
    assert e.value.code == 2

    message = str(e.value.message)
    if backend is argparse.backend:  # pyright: ignore
        assert "the following arguments are required: arg" in message.lower()
    else:
        assert message == "Argument 'arg arg' requires 2 values, found 1 ('arg' so far)"



# type: ignore
from __future__ import annotations

import attr
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse

factory = attr.Factory(lambda: [4])


@attr.s
class Command:
    name: str = attr.ib()
    default: Annotated[int, cappa.Arg(long=True)] = attr.ib(default=4)
    default_factory: Annotated[list[int], cappa.Arg(long=True)] = attr.ib(
        default=factory
    )


@backends
def test_attrs(backend: Backend):
    result = parse(Command, "meow", backend=backend)
    assert result == Command(name="meow", default=4, default_factory=[4])



from __future__ import annotations

from dataclasses import dataclass, field

import pytest

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    name: str = field(metadata={"cappa": cappa.Arg(short=True)})
    cmd: Invalid = field(metadata={"cappa": cappa.Subcommand()})


@dataclass
class Invalid:
    a: int = field(metadata={"cappa": 4})


@backends
def test_valid(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(Command, "-n", "foo", "subcommand", "4", backend=backend)

    assert "invalid" in str(e.value.message).lower()
    assert "'subcommand'" in str(e.value.message)



from __future__ import annotations

import asyncio
from contextlib import contextmanager
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke_async


@contextmanager
def sync_dep():
    yield 42


async def async_handler(value: Annotated[int, cappa.Dep(sync_dep)]):
    return value * 2


@cappa.command(invoke=async_handler)
@dataclass
class AsyncCommandWithSyncDep: ...


@backends
def test_invoke_async_with_sync_dep(backend: Backend):
    """Test that invoke_async works with a synchronous dependency."""
    result = asyncio.run(invoke_async(AsyncCommandWithSyncDep, backend=backend))
    assert result == 84



from __future__ import annotations

import typing
from dataclasses import dataclass

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    cmd: cappa.Subcommands[typing.Union[Foo, Bar]]


@dataclass
class Foo:
    f: int


@dataclass
class Bar:
    b: int


@backends
def test_subcommand_type_type_alias(backend: Backend):
    result = parse(Command, "foo", "4", backend=backend)
    assert result == Command(cmd=Foo(f=4))

    result = parse(Command, "bar", "4", backend=backend)
    assert result == Command(cmd=Bar(b=4))



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path

import cappa


@dataclass
class Todo:
    command: cappa.Subcommands[AddCommand | ListCommand]


@cappa.command(name="add")
@dataclass
class AddCommand:
    todo: str

    def __call__(self):
        path = Path("todo.json")

        data: list[str] = []
        if path.exists():
            data = json.loads(path.read_text())

        data.append(self.todo)
        path.write_text(json.dumps(data))


@cappa.command(name="list")
@dataclass
class ListCommand:
    def __call__(self, output: cappa.Output):
        path = Path("todo.json")

        data: list[str] = []
        if path.exists():
            data = json.loads(path.read_text())

        for item in data:
            output.output(f" - {item}")


cappa.invoke(Todo)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import Backend, backends, invoke, parse, strip_trailing_whitespace


def some_dep():
    return 5


@cappa.command
@dataclass
class HasExecutableMethods:
    arg: int
    include_dep: bool = False

    @cappa.command
    def add(self, some_dep: Annotated[int, cappa.Dep(some_dep)]) -> int:
        """Add two numbers."""
        if self.add:  # type: ignore[truthy-function]
            return self.arg + some_dep
        return self.arg

    @cappa.command(help="Subtract two numbers")
    def subtract(self, other: int) -> int:
        if self.add:  # type: ignore[truthy-function]
            return self.arg - other
        return self.arg


@backends
def test_parse(backend: Backend):
    result = parse(HasExecutableMethods, "10", "add", backend=backend)
    assert result == HasExecutableMethods(10, False)

    result = parse(HasExecutableMethods, "10", "subtract", "2", backend=backend)
    assert result == HasExecutableMethods(10, False)

    result = parse(HasExecutableMethods, "11", "--include-dep", "add", backend=backend)
    assert result == HasExecutableMethods(11, True)

    result = parse(
        HasExecutableMethods, "11", "--include-dep", "subtract", "2", backend=backend
    )
    assert result == HasExecutableMethods(11, True)

    with pytest.raises(Exit) as e:
        parse(HasExecutableMethods, "10", backend=backend)
    message = str(e.value.message)
    assert "required" in message
    assert "add" in message
    assert "subtract" in message


@backends
def test_invoke_add(backend: Backend):
    result = invoke(HasExecutableMethods, "10", "add", backend=backend)
    assert result == 15

    with pytest.raises(Exit) as e:
        invoke(HasExecutableMethods, "10", backend=backend)
    message = str(e.value.message)
    assert "required" in message
    assert "add" in message
    assert "subtract" in message


@backends
def test_invoke_subtract(backend: Backend):
    result = invoke(HasExecutableMethods, "10", "subtract", "7", backend=backend)
    assert result == 3

    with pytest.raises(Exit) as e:
        invoke(HasExecutableMethods, "10", backend=backend)
    message = str(e.value.message)
    assert "required" in message
    assert "add" in message
    assert "subtract" in message


@backends
def test_help(backend: Backend, capsys: Any):
    with pytest.raises(Exit):
        invoke(HasExecutableMethods, "--help", backend=backend)

    out = strip_trailing_whitespace(capsys.readouterr().out)
    assert "Add two numbers" in out
    assert "Subtract two numbers" in out


@backends
def test_nested_method_invalid_ast_source(backend: Backend, capsys: Any):
    @cappa.command
    @dataclass
    class Example:
        @cappa.command
        def add(self): ...

    with pytest.raises(Exit):
        invoke(Example, "--help", backend=backend)

    out = strip_trailing_whitespace(capsys.readouterr().out)
    assert "add" in out



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_string_group(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        name: Annotated[str, cappa.Arg(group="Strings")]

    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help", backend=backend)

    assert e.value.code == 0

    out = capsys.readouterr().out
    assert re.match(r".*Strings:?\s*\n\s*NAME.*", out, re.DOTALL)


@backends
def test_tuple_group(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        name: Annotated[str, cappa.Arg(group=(1, "Strings"))]

    with pytest.raises(cappa.HelpExit) as e:
        parse(Args, "--help", backend=backend)

    assert e.value.code == 0

    out = capsys.readouterr().out
    assert re.match(r".*Strings:?\s*\n\s*NAME.*", out, re.DOTALL)



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


async def async_parse_multiply(value: str) -> int:
    await asyncio.sleep(0.001)
    return int(value) * 2


@dataclass
class Command:
    value: Annotated[int | None, cappa.Arg(long=True, parse=async_parse_multiply)] = (
        None
    )

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    result = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert result.value is None

    result = asyncio.run(
        cappa.invoke_async(Command, argv=["--value", "10"], backend=backend)
    )
    assert result.value == 20



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Literal, Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_valid(backend: Backend):
    @dataclass
    class ArgTest:
        name: Literal["one", "two"]

    test = parse(ArgTest, "two", backend=backend)
    assert test.name == "two"


@backends
def test_valid_int(backend: Backend):
    @dataclass
    class ArgTest:
        name: Literal["one", 4]

    test = parse(ArgTest, "4", backend=backend)
    assert test.name == 4


@backends
def test_invalid(backend: Backend):
    @dataclass
    class ArgTest:
        name: Literal["one", "two", "three", 4]

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "thename", backend=backend)

    message = str(e.value.message).lower()
    assert "invalid choice: 'thename' (choose from 'one', 'two', 'three', 4)" in message


@backends
def test_unioned_literals(backend: Backend):
    @dataclass
    class ArgTest:
        name: Union[Literal["one"], Literal["two"], Literal["three"], Literal[4]]

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "thename", backend=backend)

    message = str(e.value.message).lower()
    assert (
        "invalid value for 'name': possible variants\n - literal['one']: invalid choice: 'thename' (choose from 'one')\n - literal['two']: invalid choice: 'thename' (choose from 'two')\n - literal['three']: invalid choice: 'thename' (choose from 'three')\n - literal[4]: invalid choice: 'thename' (choose from 4)"
        == message
    )


@backends
def test_invalid_collection_of_literals(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[
            set[Literal["one", "two"]] | list[int] | None, cappa.Arg(short=True)
        ] = None

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "-f", "three", backend=backend)

    assert e.value.code == 2

    err = textwrap.dedent(
        """\
        Invalid value for '-f': Possible variants
         - set[Literal['one', 'two']]: Invalid choice: 'three' (choose from 'one', 'two')
         - list[int]: invalid literal for int() with base 10: 'three'
         - <no value>"""
    )
    assert str(e.value.message).lower() == err.lower()


@backends
def test_literal_parse(backend: Backend):
    @dataclass
    class LiteralParse:
        log_level: Annotated[
            Literal["TRACE", "DEBUG", "INFO"],
            cappa.Arg(short="-L", long=True, parse=str.upper),
        ] = "INFO"

    result = parse(LiteralParse, "--log-level=debug", backend=backend)
    assert result == LiteralParse(log_level="DEBUG")



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Sc1: ...


@dataclass
class Sc2: ...


@dataclass
class Command:
    subcommand: Annotated[Union[Sc1, Sc2], cappa.Subcommand(types=[Sc1, Sc2])]


@backends
def test_required_missing(backend: Backend):
    parse(Command, "sc1", backend=backend)
    parse(Command, "sc2", backend=backend)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    subcmd: Annotated[Union[A, None], cappa.Subcommand(required=True)] = None


@dataclass
class A: ...


@backends
def test_required_implicit(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(Command)
    assert e.value.code == 2
    assert "A command is required: {[cappa.subcommand]a[/cappa.subcommand]}" in str(
        e.value.message
    )

    result = parse(Command, "a", backend=backend)
    assert result == Command(subcmd=A())



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from cappa.default import ValueFrom
from tests.utils import Backend, backends


async def async_value_with_deps(state: cappa.State[dict[str, str]]) -> str:
    await asyncio.sleep(0.001)
    return f"state-value-{state.get('key', default='default')}"


@dataclass
class Command:
    value: Annotated[str, cappa.Arg(default=ValueFrom(async_value_with_deps))]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    state: cappa.State[dict[str, str]] = cappa.State()
    state.set("key", "injected")

    result = asyncio.run(
        cappa.invoke_async(Command, argv=[], backend=backend, state=state)
    )
    assert result.value == "state-value-injected"



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    a: Annotated[str, cappa.Arg(short=True)]
    b: Annotated[str, cappa.Arg(short=True, required=True)] = "asdf"


@backends
def test_required_implicit(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(Command, "-b", "b", backend=backend)

    assert e.value.code == 2
    assert "are required: -a" in str(e.value.message)

    result = parse(Command, "-a", "a", "-b", "b", backend=backend)
    assert result == Command(a="a", b="b")


@backends
def test_required_explicit(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(Command, "-a", "a", backend=backend)

    assert e.value.code == 2
    assert "are required: -b" in str(e.value.message)

    result = parse(Command, "-a", "a", "-b", "b")
    assert result == Command(a="a", b="b")


@backends
def test_required_lists_all(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(Command, backend=backend)

    assert e.value.code == 2
    assert "are required: -a, -b" in str(e.value.message)


@backends
def test_required_explicit_false(backend: Backend):
    @dataclass
    class Example:
        a: Annotated[str, cappa.Arg(short="-a", required=False)]
        c: Annotated[str, cappa.Arg(short="-c", required=False)]

    with pytest.raises(ValueError) as e:
        parse(Example, "-a", "a", backend=backend)

    assert (
        "When specifying `required=False`, a default value must be supplied able to be "
        "supplied through type inference, `Arg(default=...)`, or through class-level default"
    ) == str(e.value)


@backends
def test_required_explicit_false_union_none_no_explicit_default(backend: Backend):
    @dataclass
    class Example:
        c: Annotated[Union[str, None], cappa.Arg(short="-c", required=False)]

    result = parse(Example, backend=backend)
    assert result == Example(c=None)

    result = parse(Example, "-c", "c", backend=backend)
    assert result == Example(c="c")


@backends
def test_required_unbounded_list(backend: Backend):
    @dataclass
    class Example:
        c: Annotated[list[str], cappa.Arg(required=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Example, backend=backend)
    assert "require" in str(e.value.message)

    result = parse(Example, "c", backend=backend)
    assert result == Example(["c"])


@backends
def test_required_option(backend: Backend):
    @dataclass
    class Example:
        c: Annotated[list[str], cappa.Arg(short=True, required=True)]

    with pytest.raises(cappa.Exit) as e:
        parse(Example, backend=backend)
    assert "the following arguments are required: -c" == str(e.value.message).lower()

    result = parse(Example, "-c", "c", backend=backend)
    assert result == Example(["c"])



import sys
from dataclasses import dataclass
from typing import Any

import pytest
from rich.console import Console
from rich.text import Text

import cappa
from tests.utils import Backend, backends, parse


@pytest.mark.help
@backends
def test_arg_renders_text(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        foo: str
        """[yellow]This is yellow[/]"""

    with pytest.raises(cappa.Exit):
        parse(
            Args,
            "--help",
            backend=backend,
            help_formatter=cappa.HelpFormatter().with_arg_format(
                Text("[red]Help:[/] {help}")
            ),
            output=cappa.Output(
                output_console=Console(force_terminal=True, file=sys.stdout)
            ),
        )

    result = capsys.readouterr()

    red = "\x1b[31m"
    yellow = "\x1b[33m"
    end = "\x1b[0m"

    expected_result = f"{red}Help:{end} {yellow}This is yellow{end}"
    assert expected_result in result.out

    assert str(Text.from_ansi(expected_result)) == "Help: This is yellow"


@pytest.mark.help
@backends
def test_arg_renders_string(backend: Backend, capsys: Any):
    @dataclass
    class Args:
        foo: str
        """[yellow]This is yellow[/]"""

    with pytest.raises(cappa.Exit):
        parse(
            Args,
            "--help",
            backend=backend,
            help_formatter=cappa.HelpFormatter(arg_format="Help: {help}"),
            output=cappa.Output(
                output_console=Console(force_terminal=True, file=sys.stdout)
            ),
        )

    result = capsys.readouterr()

    yellow = "\x1b[33m"
    end = "\x1b[0m"

    expected_result = f"Help: {yellow}This is yellow{end}"
    assert expected_result in result.out

    assert str(Text.from_ansi(expected_result)) == "Help: This is yellow"



from __future__ import annotations

import contextlib
import io
import logging
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke

log = logging.getLogger("test")


@contextlib.contextmanager
def binary_io():
    with io.BytesIO() as buffer:
        yield buffer


def command(
    binary_io: Annotated[io.BytesIO, cappa.Dep(binary_io)],
):
    assert not binary_io.closed
    binary_io.write(b"hello")
    return binary_io, binary_io.getvalue()


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(backend: Backend):
    buffer, result = invoke(Command, backend=backend)

    assert buffer.closed is True
    assert result == b"hello"



from __future__ import annotations

import typing
from collections.abc import Iterable
from dataclasses import dataclass, replace
from itertools import groupby
from typing import Any, List, Sequence, Tuple, cast

from rich.console import Console, NewLine
from rich.markdown import Markdown
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from typing_extensions import Self, TypeAlias

from cappa.arg import Arg, ArgAction, FinalArg
from cappa.output import Displayable
from cappa.subcommand import FinalSubcommand

if typing.TYPE_CHECKING:
    from cappa.command import FinalCommand

Dimension: TypeAlias = typing.Tuple[int, int, int, int]


@dataclass(frozen=True)
class FieldGroup:
    """A group of arguments or a subcommand sharing the same field_name.

    Either `args` or `subcommand` will be set, never both.
    """

    field_name: str | None
    required: bool
    args: list[FinalArg[Any]] | None = None
    subcommand: FinalSubcommand | None = None


@dataclass(frozen=True)
class ArgGroup:
    """A group of arguments/subcommands organized by their Group."""

    name: str
    field_groups: list[FieldGroup]

    @staticmethod
    def _by_group_key(
        arg: FinalArg[Any] | FinalSubcommand,
    ) -> tuple[int, int, str, bool]:
        return arg.group.key

    @staticmethod
    def _by_group(arg: FinalArg[Any] | FinalSubcommand) -> str:
        return arg.group.name

    @classmethod
    def collect(
        cls, command: FinalCommand[Any], include_hidden: bool = False
    ) -> list[ArgGroup]:
        """Collect and group arguments from a command by their Group."""
        sorted_args = sorted(command.all_arguments, key=cls._by_group_key)

        result: list[ArgGroup] = []
        for name, args in groupby(sorted_args, key=cls._by_group):
            items = [a for a in args if include_hidden or not a.hidden]

            field_grouped: dict[str | None, list[FinalArg[Any] | FinalSubcommand]] = {}
            for item in items:
                field_name = item.field_name or None
                if field_name not in field_grouped:
                    field_grouped[field_name] = []
                field_grouped[field_name].append(item)

            field_groups: list[FieldGroup] = []
            for field_name, items in field_grouped.items():
                # Validate that items are all Args or all Subcommands
                first_item = items[0]
                if isinstance(first_item, FinalArg):
                    # Validate all are Args
                    if not all(isinstance(item, FinalArg) for item in items):
                        raise ValueError(  # pragma: no cover
                            f"FieldGroup with field_name={field_name!r} contains mixed types: "
                            f"Args and Subcommands cannot share the same field_name"
                        )
                    group_args = cast(List[FinalArg[Any]], items)

                    # All items must have the same requiredness
                    required_values = {arg.required for arg in group_args}
                    assert len(required_values) == 1, (
                        f"All items with field_name={field_name!r} must have the same "
                        f"required value, got: {required_values}"
                    )
                    required = required_values.pop()
                    field_groups.append(
                        FieldGroup(field_name, required, args=group_args)
                    )
                else:
                    assert len(items) == 1
                    assert isinstance(first_item, FinalSubcommand)
                    field_groups.append(
                        FieldGroup(
                            field_name,
                            first_item.required,
                            subcommand=first_item,
                        )
                    )

            result.append(cls(name=name, field_groups=field_groups))

        return result


TextComponent = typing.Union[Text, Markdown, str]
ArgFormat: TypeAlias = typing.Union[
    TextComponent,
    typing.Callable[[Arg[Any]], typing.Union[TextComponent, None]],
]
ArgFormats: TypeAlias = typing.Sequence[ArgFormat]


class HelpFormattable(typing.Protocol):
    left_padding: Dimension
    arg_format: ArgFormat | ArgFormats
    default_format: str

    def long(
        self, command: FinalCommand[Any], prog: str
    ) -> list[Displayable]: ...  # pragma: no cover

    def short(
        self, command: FinalCommand[Any], prog: str
    ) -> Displayable: ...  # pragma: no cover


@dataclass(frozen=True)
class HelpFormatter(HelpFormattable):
    left_padding: Dimension = (0, 0, 0, 2)
    arg_format: ArgFormat | ArgFormats = (
        Markdown("{help}"),
        Markdown("{choices}"),
        Markdown("{default}", style="dim italic"),
    )
    default_format: str = "(Default: {default})"

    default: typing.ClassVar[HelpFormatter]

    def long(self, command: FinalCommand[Any], prog: str) -> list[Displayable]:
        arg_groups = ArgGroup.collect(command)

        lines: list[Displayable] = []
        lines.append(add_short_args(prog, arg_groups))

        if command.help:
            lines.append(NewLine())
            lines.append(Padding(Markdown(f"**{command.help}**"), self.left_padding))
        if command.description:
            lines.append(NewLine())
            lines.append(Padding(Markdown(command.description), self.left_padding))

        console = Console()
        lines.extend(add_long_args(console, self, arg_groups))

        if command.epilog:
            lines.append(NewLine())
            lines.append(Padding(Markdown(command.epilog), self.left_padding))

        return lines

    def short(self, command: FinalCommand[Any], prog: str) -> Displayable:
        arg_groups = ArgGroup.collect(command)
        return add_short_args(prog, arg_groups)

    def with_arg_format(
        self, _format: ArgFormat | tuple[ArgFormat, ...], *formats: ArgFormat
    ) -> Self:
        format = cast(
            Tuple[ArgFormat, ...], _format if isinstance(_format, tuple) else (_format,)
        )
        arg_format = (*format, *formats)
        return replace(self, arg_format=arg_format)

    def with_default_format(self, format: str) -> Self:
        return replace(self, default_format=format)


HelpFormatter.default = HelpFormatter()


def add_long_args(
    console: Console, help_formatter: HelpFormatter, arg_groups: list[ArgGroup]
) -> list[Table]:
    table = Table(box=None, expand=False, padding=help_formatter.left_padding)
    table.add_column(justify="left", ratio=1)
    table.add_column(style="cappa.help", ratio=2)

    for group in arg_groups:
        table.add_row(
            Text(group.name, style="cappa.group", justify="left"),
            Text(style="cappa.group"),
        )

        for field_group in group.field_groups:
            if field_group.args:
                combined = format_arg_name(field_group, ", ")

                help_text = format_args(console, help_formatter, *field_group.args)

                table.add_row(
                    Padding(combined, help_formatter.left_padding),
                    help_text,
                )

            else:
                assert field_group.subcommand
                subcommand = field_group.subcommand
                for option in subcommand.available_options():
                    table.add_row(
                        *format_subcommand(help_formatter, option, subcommand)
                    )

        table.add_row()

    return [table]


def format_args(
    console: Console, help_formatter: HelpFormattable, *args: FinalArg[Any]
) -> Displayable:
    """Format multiple args with the same field_name by concatenating their help texts."""
    segments: list[TextComponent] = []

    for arg in args:
        unknown_arg_format = help_formatter.arg_format
        if isinstance(unknown_arg_format, Iterable) and not isinstance(
            unknown_arg_format, str
        ):
            arg_format = cast(Sequence[ArgFormat], unknown_arg_format)
        else:
            arg_format = (unknown_arg_format,)

        for format_segment in arg_format:
            default = arg.show_default.format_default(
                arg.default, help_formatter.default_format
            )

            choices = ""
            if arg.choices:
                choices = "Valid options: " + ", ".join(arg.choices) + "."

            context: dict[str, Any] = {
                "help": arg.help or "",
                "default": default,
                "choices": choices,
                "arg": arg,
            }

            if callable(format_segment):
                format_segment = cast(TextComponent, format_segment(arg))

            format_segment_text = _get_text_component_text(format_segment)
            if not format_segment_text:
                continue

            formatted_text = format_segment_text.format(**context)
            segment = _replace_rich_text_component(format_segment, formatted_text)

            if segment:
                segments.append(segment)

    return _markdown_to_text(console, segments)


def _markdown_to_text(console: Console, renderables: Sequence[TextComponent]) -> Text:
    # Render the text without line-wrapping because we're not yet printing it and it should reflow
    # until the point at which it's being printed to the user.
    render_console = Console(soft_wrap=True)
    render_options = render_console.options.update(no_wrap=True, overflow="ignore")

    result = Text()
    for renderable in renderables:
        if isinstance(renderable, Markdown):
            renderable_result = Text()
            for segment in render_console.render(renderable, render_options):
                text = segment.text
                if not text:
                    continue
                if text.startswith(" "):  # dedup leading spaces
                    text = " " + text.lstrip()
                if text.endswith(" "):  # dedup trailing spaces
                    text = text.rstrip() + " "
                renderable_result.append(Text(text, style=segment.style or "", end=""))
            renderable_result.rstrip()
            if renderable_result:
                _append_to_text(result, renderable_result)
        else:
            if isinstance(renderable, str):
                renderable = Text.from_markup(renderable)

            _append_to_text(result, renderable)

    return result


def _append_to_text(text: Text, component: Text) -> None:
    # When a component is multiline (like long, structured {help}), it looks weird when components
    # like {choices} or {default} are appended to the end of the last line. In such cases, add a newline.
    # Sort of a weird heuristic, we'll see if this holds well in practice.
    if text:
        text.append("\n" if "\n" in text.plain else " ")
    text.append(component)


def _get_text_component_text(c: TextComponent) -> str:
    if isinstance(c, Text):
        return c.plain

    if isinstance(c, Markdown):
        return c.markup

    return c


def _replace_rich_text_component(c: TextComponent, text: str) -> TextComponent:
    if isinstance(c, Text):
        return Text.from_markup(
            text,
            style=c.style,
            justify=c.justify,
            overflow=c.overflow,
            end=c.end,
        )

    if isinstance(c, Markdown):
        return Markdown(
            text,
            code_theme=c.code_theme,
            justify=c.justify,
            style=c.style,
            hyperlinks=c.hyperlinks,
            inline_code_lexer=c.inline_code_lexer,
            inline_code_theme=c.inline_code_theme,
        )

    return text


def format_subcommand(
    help_formatter: HelpFormatter,
    command: FinalCommand[Any],
    subcommand: FinalSubcommand | None = None,
):
    canonical = command.real_name()
    parts = [f"[cappa.subcommand]{canonical}[/cappa.subcommand]"]
    if subcommand is not None:
        for alias in subcommand.visible_aliases_for(canonical):
            label = alias.name
            if alias.deprecated:
                label = f"{label} (deprecated)"
            parts.append(f"[cappa.subcommand]{label}[/cappa.subcommand]")
    return (
        Padding(", ".join(parts), help_formatter.left_padding),
        command.help,
    )


def add_short_args(prog: str, arg_groups: list[ArgGroup]) -> str:
    segments: list[str] = [f"Usage: {prog}"]
    for group in arg_groups:
        for field_group in group.field_groups:
            segments.append(format_arg_name(field_group, ", ", n=1))

    return " ".join(segments)


def format_arg_name(item: FieldGroup, delimiter: str, *, n: int = 0) -> str:
    """Format the name(s) of args/subcommands for display.

    If given a FieldGroup, formats all items together with appropriate brackets.
    If given a single Arg/Subcommand, formats just that item.
    """
    if item.args:
        parts: list[str] = []
        for arg in item.args:
            num_args = arg.num_args
            has_value = (
                not ArgAction.is_non_value_consuming(arg.action) and num_args.n != 0
            )

            arg_names = arg.names_str(delimiter, n=n)
            if not arg.is_option:
                arg_names = arg_names.upper()

                if num_args.n == -1:
                    arg_names = f"{arg_names} ..."

            text = f"[cappa.arg]{arg_names}[/cappa.arg]"

            if arg.is_option and has_value:
                name = arg.value_name.upper()
                if num_args.n == -1:
                    name = f"{name} ..."

                text = f"{text} [cappa.arg.name]{name}[/cappa.arg.name]"

            parts.append(text)
        combined = ", ".join(parts)

        if not item.required:
            return f"[{combined}]"
        return combined

    assert item.subcommand is not None
    arg_names = item.subcommand.names_str(",")
    return f"{{[cappa.subcommand]{arg_names}[/cappa.subcommand]}}"


def format_subcommand_names(names: list[str]):
    return ", ".join(f"[cappa.subcommand]{a}[/cappa.subcommand]" for a in names)



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def two():
    return 2


def three(_: Command):
    return 3


def one(two: Annotated[int, cappa.Dep(two)], three: Annotated[int, cappa.Dep(three)]):
    return (two, three)


def command(one: Annotated[int, cappa.Dep(one)]):
    return one


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(backend: Backend):
    def stub_two():
        return 4

    result = invoke(Command, deps={two: cappa.Dep(stub_two), three: 6}, backend=backend)
    assert result == (4, 6)



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_missing_default(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[bool, cappa.Arg(long="--foo/--no-foo")] = False

    test = parse(Args, backend=backend)
    assert test.foo is False

    test = parse(Args, "--foo", backend=backend)
    assert test.foo is True

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--foo", "--no-foo", backend=backend)

    message = str(e.value.message)
    assert "--no-foo" in message
    assert "--foo" in message
    assert "not allowed with argument" in message



from __future__ import annotations

from dataclasses import dataclass

import cappa


@dataclass
class DeferImport:
    command: cappa.Subcommands[MeanCommand]


@cappa.command(name="mean", invoke="defer_import.mean.calculate_mean")
@dataclass
class MeanCommand:
    numbers: list[int]


def main():
    cappa.invoke(DeferImport)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.arg import Group
from tests.utils import Backend, backends, parse


@backends
def test_implicit_syntax(backend: Backend | None):
    @dataclass
    class ArgTest:
        verbose: Annotated[
            int,
            cappa.Arg(short="-v", action=cappa.ArgAction.count),
            cappa.Arg(long="--verbosity"),
        ] = 0

    result = parse(ArgTest, backend=backend)
    assert result.verbose == 0

    result = parse(ArgTest, "-vv", backend=backend)
    assert result.verbose == 2

    result = parse(ArgTest, "--verbosity", "3", backend=backend)
    assert result.verbose == 3

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "-vv", "--verbosity", "3", backend=backend)
    if backend is None:
        message = "Argument '--verbosity' is not allowed with argument '-v'"
    else:
        message = "argument --verbosity: not allowed with argument -v"

    assert message.lower() == str(e.value.message).lower()


@backends
def test_implicit_syntax_explicit_name(backend: Backend, capsys: Any):
    @dataclass
    class ArgTest:
        verbose: Annotated[
            int,
            cappa.Arg(short="-v", action=cappa.ArgAction.count, group="Foo"),
            cappa.Arg(long="--verbosity", group="Foo"),
        ] = 0

    with pytest.raises(cappa.Exit):
        parse(ArgTest, "-h")

    out = capsys.readouterr().out

    assert "[-v, --verbosity VERBOSE]" in out
    assert "(Default: 0)" in out


@backends
def test_explicit_groups(backend: Backend | None):
    @dataclass
    class ArgTest:
        verbose: Annotated[
            int,
            cappa.Arg(
                short="-v",
                action=cappa.ArgAction.count,
                group=Group(exclusive=True),
            ),
        ] = 0
        verbosity: Annotated[
            int,
            cappa.Arg(long="--verbosity", group=Group(exclusive=True)),
        ] = 0

    result = parse(ArgTest, backend=backend)
    assert result.verbose == 0
    assert result.verbosity == 0

    result = parse(ArgTest, "-vv", backend=backend)
    assert result.verbose == 2
    assert result.verbosity == 0

    result = parse(ArgTest, "--verbosity", "3", backend=backend)
    assert result.verbose == 0
    assert result.verbosity == 3

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "-vv", "--verbosity", "3", backend=backend)

    if backend is None:
        message = "Argument '--verbosity' is not allowed with argument '-v'"
    else:
        message = "argument --verbosity: not allowed with argument -v"

    assert message.lower() == str(e.value.message).lower()


@backends
def test_differing_group_identity(backend: Backend):
    @dataclass
    class ArgTest:
        one: Annotated[int, cappa.Arg(group=Group(name="Verbose", exclusive=False))] = 0
        tw: Annotated[int, cappa.Arg(group=Group(name="Verbose", exclusive=True))] = 0

    with pytest.raises(ValueError) as e:
        parse(ArgTest, backend=backend)

    assert str(e.value) == (
        "Group details do not match among arguments: Group(id='Verbose', exclusive=True) != Group(id='Verbose', exclusive=False)."
    )



from __future__ import annotations

import asyncio
import contextlib
import io
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


@contextlib.contextmanager
def binary_io():
    with io.BytesIO() as buffer:
        yield buffer


def command(
    binary_io: Annotated[io.BytesIO, cappa.Dep(binary_io)],
):
    assert not binary_io.closed
    binary_io.write(b"hello from command")
    return binary_io


@cappa.command(invoke=command)
@dataclass
class Command:
    pass


@backends
def test(backend: Backend):
    buffer = asyncio.run(cappa.invoke_async(Command, argv=[], backend=backend))
    assert buffer.closed



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class CollisionSub:
    field1: str = ""
    field2: Annotated[str, cappa.Arg(long=True)] = ""


@dataclass
class Collision:
    sub: cappa.Destructured[CollisionSub]
    field1: int = 0
    field2: int = 0


def test_destructured_collision():
    result = parse(Collision, "1")
    assert result == Collision(sub=CollisionSub("1", ""), field1=0)

    result = parse(Collision, "1", "2")
    assert result == Collision(sub=CollisionSub("1", ""), field1=2)

    result = parse(Collision, "--field2=-1", "1", "2", "3")
    assert result == Collision(sub=CollisionSub("1", "-1"), field1=2, field2=3)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def top_level_command(tlc: TopLevelCommand):
    return ("tlc", tlc)


def sub_command(sub: Subcommand):
    return ("sub", sub)


@cappa.command(name="sub", invoke=sub_command)
@dataclass
class Subcommand:
    bar: int


@cappa.command(invoke=top_level_command)
@dataclass
class TopLevelCommand:
    foo: Annotated[int, cappa.Arg(long=True)] = 4
    subcommand: Annotated[Union[Subcommand, None], cappa.Subcommand] = None


@backends
def test_invoke_top_level_command(backend: Backend):
    result = invoke(TopLevelCommand, "--foo", "4", backend=backend)
    assert result == ("tlc", TopLevelCommand(foo=4))


@backends
def test_invoke_subcommand(backend: Backend):
    result = invoke(TopLevelCommand, "sub", "4", backend=backend)
    assert result == ("sub", Subcommand(bar=4))



import textwrap
from dataclasses import dataclass
from typing import Any

import pytest
from rich.table import Table
from rich.text import Text
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_invoke_exit_success_with_message(capsys: Any, backend: Backend):
    def fn():
        raise cappa.Exit("With message")

    @cappa.command(invoke=fn)
    @dataclass
    class Example: ...

    with pytest.raises(cappa.Exit):
        cappa.invoke(Example, argv=[], backend=backend)

    out = capsys.readouterr().out
    assert out == "With message\n"


@backends
def test_invoke_exit_success_without_message(capsys: Any, backend: Backend):
    def fn():
        raise cappa.Exit()

    @cappa.command(invoke=fn)
    @dataclass
    class Example: ...

    with pytest.raises(cappa.Exit):
        cappa.invoke(Example, argv=[], backend=backend)

    out = capsys.readouterr().out
    assert out == ""


@backends
def test_invoke_exit_error(capsys: Any, backend: Backend):
    def fn():
        raise cappa.Exit("With message", code=1)

    @cappa.command(invoke=fn)
    @dataclass
    class Example: ...

    with pytest.raises(cappa.Exit) as e:
        cappa.invoke(Example, argv=[], backend=backend)

    assert e.value.code == 1
    out = capsys.readouterr().err
    assert out == "Error: With message\n"


@backends
def test_error_output_rich_text(capsys: Any, backend: Backend):
    def fn():
        raise cappa.Exit(Text("With message"), code=1)

    @cappa.command(invoke=fn)
    @dataclass
    class Example: ...

    with pytest.raises(cappa.Exit) as e:
        cappa.invoke(Example, argv=[], backend=backend)

    assert e.value.code == 1
    out = capsys.readouterr().err
    assert out == "Error: With message\n"


@backends
def test_explicit_output_prefix(capsys: Any, backend: Backend):
    @cappa.command(name="asdf")
    @dataclass
    class Example: ...

    output = cappa.Output(error_format="{prog}: error({code}): {message}.")
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "--fooooo", backend=backend, output=output)

    assert e.value.code == 2
    out = capsys.readouterr().err
    assert out.lower() == "asdf: error(2): unrecognized arguments: --fooooo.\n"


def _debug(output: cappa.Output):
    table = Table(style="blue")
    table.add_row("one", "two")
    output.error(table)
    raise cappa.Exit(code=0)


@backends
def test_output_formatting_complex_rich_object(capsys: Any, backend: Backend):
    @dataclass
    class Example:
        debug_info: Annotated[
            bool,
            cappa.Arg(long=True, action=_debug, num_args=0),
        ] = False

    output = cappa.Output(error_format="[red]Error[/red]:\n{message}!")
    with pytest.raises(cappa.Exit) as e:
        parse(Example, "--debug-info", output=output, backend=backend)

    assert e.value.code == 0
    out = capsys.readouterr().err
    assert out == textwrap.dedent(
        """\
        Error:
        ┏━━━━━┳━━━━━┓
        ┃     ┃     ┃
        ┡━━━━━╇━━━━━┩
        │ one │ two │
        └─────┴─────┘!
        """
    )
